# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Grouped matrix multiplication kernels for TPU written in Pallas."""

import functools
from collections.abc import Callable
from typing import Any, Optional

import jax
import jax.numpy as jnp
from jax import lax
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu

from tpu_inference.kernels.megablox import common

partial = functools.partial


def _validate_args(
    *,
    lhs: jnp.ndarray,
    rhs: jnp.ndarray,
    group_sizes: jnp.ndarray,
    rhs_scale: jnp.ndarray | None = None,
    rhs_bias: jnp.ndarray | None = None,
):
    """Validates the arguments for the gmm function."""
    # Validate 'lhs'.
    if lhs.ndim != 2:
        raise ValueError(f"Expected 2-tensor for 'lhs' but got {lhs.ndim=}.")
    common.assert_is_supported_dtype(lhs.dtype)

    # Validate 'rhs'.
    if rhs.ndim != 3:
        raise ValueError(f"Expected 3-tensor for 'rhs' but got {rhs.ndim=}.")
    common.assert_is_supported_dtype(rhs.dtype)

    if lhs.shape[1] != rhs.shape[1]:
        raise ValueError(
            "Expected 'lhs' and 'rhs' to have the same number of input features."
            f" But instead got {lhs.shape[1]=} and {rhs.shape[1]=}")

    # Validate 'group_sizes'.
    if group_sizes.dtype != jnp.int32:
        raise ValueError(
            f"Expected 32-bit integer 'group_sizes' but got {group_sizes.dtype=}."
        )

    num_groups, in_size, out_size = rhs.shape

    if rhs_scale is not None:
        # Validate 'rhs_scale'.
        if rhs_scale.ndim != 4:
            raise ValueError(
                f"Expected 4-tensor for 'rhs_scale' but got {rhs_scale.ndim=}."
            )
        expected_rhs_scale_shape = (num_groups, rhs_scale.shape[1], 1,
                                    out_size)
        if rhs_scale.shape != expected_rhs_scale_shape:
            raise ValueError(
                "Expected 'rhs_scale' to have the shape of"
                f" {expected_rhs_scale_shape} but got {rhs_scale.shape=}.")

    if rhs_bias is not None:
        # Validate 'rhs_bias'.
        if rhs_bias.ndim != 3:
            raise ValueError(
                f"Expected 3-tensor for 'rhs_bias' but got {rhs_bias.ndim=}.")
        expected_rhs_bias_shape = (num_groups, 1, out_size)
        if rhs_bias.shape != expected_rhs_bias_shape:
            raise ValueError(
                "Expected 'rhs_bias' to have the shape of"
                f" {expected_rhs_bias_shape} but got {rhs_bias.shape=}.")


def _calculate_num_tiles(x: int, tx: int) -> int:
    tiles, rem = divmod(x, tx)
    if rem:
        raise ValueError(
            f"{x} must be divisible by x-dimension tile size ({tx}).")
    return tiles


def _calculate_irregular_num_tiles(x: int, tx: int) -> tuple[int, int]:
    tiles, rem = divmod(x, tx)
    if rem:
        tiles += 1
    return tiles, rem


GroupMetadata = Any  # TODO(enriqueps): Clean this up and use a namedtuple


def make_group_metadata(
    *,
    group_sizes: jnp.ndarray,
    m: int,
    tm: int,
    start_group: jnp.ndarray,
    num_nonzero_groups: int,
    visit_empty_groups: bool = True,
) -> GroupMetadata:
    """Create the metadata needed for grouped matmul computation.

  Args:
    group_sizes: A 1d, jnp.ndarray with shape [num_groups] and jnp.int32 dtype.
    m: The number of rows in lhs.
    tm: The m-dimension tile size being used.
    start_group: The group in group sizes to start computing from. This is
      particularly useful for when rhs num_groups is sharded.
    num_nonzero_groups: Number of groups in group sizes to compute on. Useful in
      combination with group_offset.
    visit_empty_groups: If True, do not squeeze tiles for empty groups out of
      the metadata. This is necessary for tgmm, where we at least need to zero
      the output for each group.

  Returns:
    tuple of:
      group_offsets: A 1d, jnp.ndarray with shape [num_groups+1] and jnp.int32
        dtype. group_offsets[i] indicates the row at which group [i] starts in
        the lhs matrix and group_offsets[i-1] = m.
      group_ids: A 1d, jnp.ndarray with shape [m_tiles + num_groups] and
        jnp.int32 dtype. group_ids[i] indicates which group grid index 'i' will
        work on.
      m_tile_ids: A 1d, jnp.ndarray with shape [m_tiles + num_groups] and
        jnp.int32. m_tile_ids[i] indicates which m-dimension tile grid index 'i'
        will work on.
    num_tiles: The number of m-dimension tiles to execute.
  """
    num_groups = group_sizes.shape[0]
    end_group = start_group + num_nonzero_groups - 1

    # Calculate the offset of each group, starting at zero. This metadata is
    # similar to row offsets in a CSR matrix. The following properties hold:
    #
    # group_offsets.shape = [num_groups + 1]
    # group_offsets[0] = 0
    # group_offsets[num_groups] = m
    #
    # The row at which group 'i' starts is group_offsets[i].
    group_ends = jnp.cumsum(group_sizes)
    group_offsets = jnp.concatenate(
        [jnp.zeros(1, dtype=jnp.int32), group_ends])

    # Assign a group id to each grid index.
    #
    # If a group starts somewhere other than the start of a tile or ends somewhere
    # other than the end of a tile we need to compute that full tile. Calculate
    # the number of tiles for each group by rounding their end up to the nearest
    # 'tm' and their start down to the nearest 'tm'.

    # (1) Round the group_ends up to the nearest multiple of 'tm'.
    #
    # NOTE: This does not change group_offsets[num_groups], which is m
    # (because we enforce m is divisible by tm).
    rounded_group_ends = ((group_ends + tm - 1) // tm * tm).astype(jnp.int32)

    # (2) Round the group_starts down to the nearest multiple of 'tm'.
    group_starts = jnp.concatenate(
        [jnp.zeros(1, dtype=jnp.int32), group_ends[:-1]])
    rounded_group_starts = group_starts // tm * tm

    # (3) Calculate the number of rows in each group.
    #
    # NOTE: Handle zero-sized groups as a special case. If the start for a
    # zero-sized group is not divisible by 'tm' its start will be rounded down and
    # its end will be rounded up such that its size will become 1 tile here.
    rounded_group_sizes = rounded_group_ends - rounded_group_starts
    rounded_group_sizes = jnp.where(group_sizes == 0, 0, rounded_group_sizes)

    # (4) Convert the group sizes from units of rows to unit of 'tm' sized tiles.
    #
    # An m-dimension tile is 'owned' by group 'i' if the first row of the tile
    # belongs to group 'i'. In addition to owned tiles, each group can have 0 or 1
    # initial partial tiles if it's first row does not occur in the first row of a
    # tile. The '0-th' group never has a partial tile because it always starts at
    # the 0-th row.
    #
    # If no group has a partial tile, the total number of tiles is equal to
    # 'm // tm'. If every group has a partial except the 0-th group, the total
    # number of tiles is equal to 'm // tm + num_groups - 1'. Thus we know that
    #
    # tiles_m <= group_tiles.sum() <= tiles_m + num_groups - 1
    #
    # Where tiles_m = m // tm.
    #
    # NOTE: All group sizes are divisible by 'tm' because of the rounding in steps
    # (1) and (2) so this division is exact.
    group_tiles = rounded_group_sizes // tm

    if visit_empty_groups:
        # Insert one tile for empty groups.
        group_tiles = jnp.where(group_sizes == 0, 1, group_tiles)

    # Create the group ids for each grid index based on the tile counts for each
    # group.
    #
    # NOTE: This repeat(...) will pad group_ids with the final group id if
    # group_tiles.sum() < tiles_m + num_groups - 1. The kernel grid will be sized
    # such that we only execute the necessary number of tiles.
    tiles_m = _calculate_num_tiles(m, tm)
    group_ids = jnp.repeat(
        jnp.arange(num_groups, dtype=jnp.int32),
        group_tiles,
        total_repeat_length=tiles_m + num_groups - 1,
    )

    # Assign an m-dimension tile id to each grid index.
    #
    # NOTE: Output tiles can only be re-visited consecutively. The following
    # procedure guarantees that m-dimension tile indices respect this.

    # (1) Calculate how many times each m-dimension tile will be visited.
    #
    # Each tile is guaranteed to be visited once by the group that owns the tile.
    # The remaining possible visits occur when a group starts inside of a tile at
    # a position other than the first row. We can calculate which m-dimension tile
    # each group starts in by floor-dividing its offset with `tm` and then count
    # tile visits with a histogram.
    #
    # To avoid double counting tile visits from the group that owns the tile,
    # filter these out by assigning their tile id to `tile_m` (one beyond the max)
    # such that they're ignored by the subsequent histogram. Also filter out any
    # group which is empty.
    #
    # TODO(tgale): Invert the 'partial_tile_mask' predicates to be more clear.
    partial_tile_mask = jnp.logical_or((group_offsets[:-1] % tm) == 0,
                                       group_sizes == 0)

    # Explicitly enable tiles for zero sized groups, if specified. This covers
    # zero sized groups that start on a tile-aligned row and those that do not.
    if visit_empty_groups:
        partial_tile_mask = jnp.where(group_sizes == 0, 0, partial_tile_mask)

    partial_tile_ids = jnp.where(partial_tile_mask, tiles_m,
                                 group_offsets[:-1] // tm)

    tile_visits = (jnp.histogram(
        partial_tile_ids, bins=tiles_m, range=(0, tiles_m - 1))[0] + 1)

    # Create the m-dimension tile ids for each grid index based on the visit
    # counts for each tile.
    m_tile_ids = jnp.repeat(
        jnp.arange(tiles_m, dtype=jnp.int32),
        tile_visits.astype(jnp.int32),
        total_repeat_length=tiles_m + num_groups - 1,
    )

    # Account for sharding.
    #
    # Find the start of the groups owned by our shard and shift the group_ids and
    # m_tile_ids s.t. the metadata for our tiles are at the front of the arrays.
    #
    # TODO(tgale): Move this offset into the kernel to avoid these rolls.
    first_tile_in_shard = (group_ids < start_group).sum()
    group_ids = jnp.roll(group_ids, shift=-first_tile_in_shard, axis=0)
    m_tile_ids = jnp.roll(m_tile_ids, shift=-first_tile_in_shard, axis=0)

    # Calculate the number of tiles we need to compute for our shard.
    #
    # Remove tile visits that belong to a group not in our shard.
    iota = jnp.arange(num_groups, dtype=jnp.int32)
    active_group_mask = jnp.logical_and(iota <= end_group, iota >= start_group)
    group_tiles = jnp.where(active_group_mask, group_tiles, 0)
    num_tiles = group_tiles.sum()
    return (group_offsets, group_ids, m_tile_ids), num_tiles


def _get_store_mask(
    *,
    grid_id: jnp.ndarray,
    group_metadata: GroupMetadata,
    tm: int,
    tn: int,
) -> jnp.ndarray:
    """Mask for rows that belong to the current group in the current tile."""
    group_offsets, group_ids, m_tile_ids = group_metadata[:3]
    group_id = group_ids[grid_id]
    group_start = group_offsets[group_id]
    group_end = group_offsets[group_id + 1]
    m_id = m_tile_ids[grid_id] * tm
    iota = jax.lax.broadcasted_iota(jnp.int32, (tm, tn), 0) + m_id
    return jnp.logical_and(iota >= group_start, iota < group_end)


def _zero_uninitialized_memory(
    out: jnp.ndarray,
    *,
    start_group: jnp.ndarray,
    num_nonzero_groups: int,
    group_metadata: GroupMetadata,
) -> jnp.ndarray:
    """Zero out uninitialized memory from output."""
    group_offsets = group_metadata[0]
    group_start = group_offsets[start_group]
    group_end = group_offsets[start_group + num_nonzero_groups]
    valid_mask = jax.lax.broadcasted_iota(jnp.int32, (out.shape[0], ), 0)
    valid_mask = (valid_mask >= group_start) & (valid_mask < group_end)
    return jnp.where(valid_mask[:, None], out, 0)


LutFn = Callable[[int, int, int], Optional[tuple[int, int, int]]]


@functools.partial(
    jax.jit,
    static_argnames=[
        "preferred_element_type",
        "tiling",
        "interpret",
    ],
)
def gmm(
    lhs: jnp.ndarray,
    rhs: jnp.ndarray,
    group_sizes: jnp.ndarray,
    preferred_element_type: jnp.dtype = jnp.float32,
    rhs_scale: jnp.ndarray | None = None,
    rhs_bias: jnp.ndarray | None = None,
    tiling: tuple[int, int, int] | LutFn | None = (128, 128, 128),
    group_offset: jnp.ndarray | None = None,
    existing_out: jnp.ndarray | None = None,
    interpret: bool = False,
) -> jnp.ndarray:
    """Compute lhs[sizes[i-1]:sizes[i], :] @ rhs for each group 'i'.

  Args:
    lhs: A 2d, jnp.ndarray with shape [m, k].
    rhs: A 3d, jnp.ndarray with shape [num_groups, n, k].
    group_sizes: A 1d, jnp.ndarray with shape [num_groups] and jnp.int32 dtype.
    preferred_element_type: jnp.dtype, the element type for the output matrix.
    rhs_scale: A 4d, jnp.ndarray with shape [num_groups, num_blocks, 1, n].
    rhs_bias: A 3d, jnp.ndarray with shape [num_groups, 1, n].
    tiling: 3-tuple of ints. The m, k and n-dimension tile sizes.
    group_offset: The group in group sizes to start computing from. This is
      particularly useful for when rhs num_groups is sharded.
    existing_out: Existing output to write to.
    interpret: Whether or not to run the kernel in interpret mode, helpful for
      testing and debugging.

  Returns:
    A 2d, jnp.ndarray with shape [m, n].
  """

    if existing_out is not None:
        assert isinstance(existing_out, jax.Array)
        expected_dtype = existing_out.dtype
        if expected_dtype != preferred_element_type:
            raise ValueError(
                "Existing output dtype must match preferred_element_type.")
    if group_offset is None:
        group_offset = jnp.array([0], dtype=jnp.int32)
    else:
        if group_offset.shape:
            raise ValueError(
                f"group_offset must be a ()-shaped array. Got: {group_offset.shape}."
            )
        group_offset = group_offset[None]
    num_current_groups = rhs.shape[0]
    num_total_groups = group_sizes.shape[0]
    _validate_args(
        lhs=lhs,
        rhs=rhs,
        group_sizes=group_sizes,
        rhs_scale=rhs_scale,
        rhs_bias=rhs_bias,
    )

    # Gather shape information.
    m, k, n = (lhs.shape[0], lhs.shape[1], rhs.shape[-1])

    # If tiling is callable, look up the problem dimensions in the LUT. If no
    # tuned tile dimensions are available throw an error.
    if callable(tiling):
        tiling = tiling(m, k, n)

    if tiling is None:
        raise ValueError(
            f"No tuned tiling found for (m, k, n) = ({m}, {k}, {n})")

    tm, tk, tn = tiling

    if rhs_scale is not None:
        assert isinstance(rhs_scale, jax.Array)
        assert rhs_scale.shape[0] == num_current_groups
        num_quant_blocks = rhs_scale.shape[1]
    else:
        num_quant_blocks = 1
    block_size = k // num_quant_blocks

    if tk > block_size or block_size % tk != 0:
        tk = block_size

    tiles_k, k_rem = _calculate_irregular_num_tiles(k, tk)
    tiles_n, n_rem = _calculate_irregular_num_tiles(n, tn)
    del n_rem

    tiles_k //= num_quant_blocks

    # Create the metadata we need for computation.
    group_metadata, num_active_tiles = make_group_metadata(  # pylint: disable=unbalanced-tuple-unpacking
        group_sizes=group_sizes,
        m=m,
        tm=tm,
        start_group=group_offset[0],
        num_nonzero_groups=rhs.shape[0],
        visit_empty_groups=False,
    )

    def kernel(
        group_metadata,
        group_offset,
        lhs,
        rhs,
        rhs_scale,
        rhs_bias,
        existing_out,
        out,
        acc_scratch,
    ):
        group_offsets, group_ids, m_tile_ids = group_metadata
        del group_offsets, group_ids, group_offset

        grid_id = pl.program_id(1)
        b_i = pl.program_id(2)
        k_i = pl.program_id(3)

        @pl.when(k_i == 0)
        def _zero_acc():
            acc_scratch[...] = jnp.zeros_like(acc_scratch)

            if existing_out is not None:
                prev_grid_id = jnp.where(grid_id > 0, grid_id - 1, 0)
                is_first_processed_group = grid_id == 0
                m_tile_changed = m_tile_ids[grid_id] != m_tile_ids[
                    prev_grid_id]
                first_time_seeing_out = jnp.logical_or(
                    is_first_processed_group, m_tile_changed)

                @pl.when(first_time_seeing_out)
                def _init_out():
                    out[...] = existing_out[...]

        def mask_k_rem(x, *, dim):
            if k_rem == 0:
                return x

            orig_dtype = x.dtype
            iota = lax.broadcasted_iota(jnp.int32, x.shape, dim)
            x = x.astype(jnp.float32)
            return jnp.where(iota < k_rem, x, 0).astype(orig_dtype)

        def _accum(is_last_k_tile, is_first_b_tile):
            if is_last_k_tile:
                mask_k_rem_lhs = partial(mask_k_rem, dim=1)
                mask_k_rem_rhs = partial(mask_k_rem, dim=0)
            else:

                def _wrapper(x):
                    return x

                mask_k_rem_lhs = _wrapper
                mask_k_rem_rhs = _wrapper

            loaded_lhs = lhs[...]
            loaded_rhs = rhs[...]

            acc = acc_scratch[...] + jnp.dot(
                mask_k_rem_lhs(loaded_lhs),
                mask_k_rem_rhs(loaded_rhs),
                preferred_element_type=jnp.float32,
            )

            if is_last_k_tile:
                if rhs_scale is not None:
                    acc *= jnp.broadcast_to(rhs_scale[...], acc.shape)

                loaded_out = out[...].astype(jnp.float32)
                if not is_first_b_tile:
                    acc += loaded_out
                elif rhs_bias is not None:
                    acc += rhs_bias[...].astype(jnp.float32)

                mask = _get_store_mask(
                    grid_id=grid_id,
                    group_metadata=group_metadata,
                    tm=tm,
                    tn=tn,
                )
                out[...] = jax.lax.select(
                    mask[...], acc, loaded_out).astype(preferred_element_type)
            else:
                acc_scratch[...] = acc

        is_last_k_tile = k_i == (tiles_k - 1)
        is_first_b_tile = b_i == 0

        lax.cond(
            is_last_k_tile,
            lambda: lax.cond(
                is_first_b_tile,
                partial(_accum, True, True),
                partial(_accum, True, False),
            ),
            partial(_accum, False, False),
        )

    def lhs_transform_indices(n_i, grid_id, b_i, k_i, group_metadata,
                              group_offset):
        # lhs is (m, k). Load the [tm, tk] matrix for this m-tile.
        group_offsets, group_ids, m_tile_ids = group_metadata
        del n_i, group_offsets, group_ids, group_offset
        return m_tile_ids[grid_id], b_i * tiles_k + k_i

    def rhs_transform_indices(n_i, grid_id, b_i, k_i, group_metadata,
                              group_offset):
        # rhs is (num_groups, k, n). Load the [tk, tn] matrix based on the group id
        # for this m-tile.
        group_offsets, group_ids, m_tile_ids = group_metadata
        del group_offsets, m_tile_ids

        # NOTE: If we're working on only a shard of the rhs we need to adjust the
        # group index we load from to account for this. The group_ids are in the
        # "unsharded" domain.
        return group_ids[grid_id] - group_offset[0], b_i * tiles_k + k_i, n_i

    def rhs_scale_transform_indices(n_i, grid_id, b_i, k_i, group_metadata,
                                    group_offset):
        group_offsets, group_ids, m_tile_ids = group_metadata
        del group_offsets, m_tile_ids, k_i
        return group_ids[grid_id] - group_offset[0], b_i, 0, n_i

    def rhs_bias_transform_indices(n_i, grid_id, b_i, k_i, group_metadata,
                                   group_offset):
        group_offsets, group_ids, m_tile_ids = group_metadata
        del group_offsets, m_tile_ids, k_i, b_i
        return group_ids[grid_id] - group_offset[0], 0, n_i

    def out_transform_indices(n_i, grid_id, b_i, k_i, group_metadata,
                              group_offset):
        # out is (m, n). Load the [tm, tn] matrix for this m-tile.
        group_offsets, group_ids, m_tile_ids = group_metadata
        del k_i, group_offsets, group_ids, group_offset, b_i
        return m_tile_ids[grid_id], n_i

    out_block_spec = pl.BlockSpec((tm, tn), out_transform_indices)
    if existing_out is None:
        in_out_block_spec: Any = None
        input_output_aliases = {}
    else:
        in_out_block_spec = out_block_spec
        input_output_aliases = {7: 0}

    lhs_block_spec = pl.BlockSpec((tm, tk), lhs_transform_indices)
    rhs_block_spec = pl.BlockSpec((None, tk, tn), rhs_transform_indices)

    if rhs_scale is None:
        rhs_scale_block_spec = None
    else:
        rhs_scale_block_spec = pl.BlockSpec((None, None, 1, tn),
                                            rhs_scale_transform_indices)

    if rhs_bias is None:
        rhs_bias_block_spec = None
    else:
        rhs_bias_block_spec = pl.BlockSpec((None, 1, tn),
                                           rhs_bias_transform_indices)

    lhs_bytes = lhs.size * lhs.itemsize
    rhs_bytes = (k * n) * rhs.itemsize  # We don't read all of rhs
    if rhs_scale is not None:
        rhs_bytes += (num_quant_blocks * n) * rhs_scale.itemsize
    if rhs_bias is not None:
        rhs_bytes += n * rhs_bias.itemsize
    out_bytes = (m * n) * jnp.dtype(preferred_element_type).itemsize
    max_active_tiles = group_metadata[1].size
    bytes_accessed = ((lhs_bytes * tiles_n) + (rhs_bytes * max_active_tiles) +
                      out_bytes)
    flops = 2 * m * k * n
    cost_estimate = pl.CostEstimate(flops=flops,
                                    bytes_accessed=bytes_accessed,
                                    transcendentals=0)
    call_gmm = pl.pallas_call(
        kernel,
        out_shape=jax.ShapeDtypeStruct((m, n), preferred_element_type),
        grid_spec=pltpu.PrefetchScalarGridSpec(
            num_scalar_prefetch=2,
            in_specs=[
                lhs_block_spec,
                rhs_block_spec,
                rhs_scale_block_spec,
                rhs_bias_block_spec,
                in_out_block_spec,
            ],
            out_specs=out_block_spec,
            grid=(tiles_n, num_active_tiles, num_quant_blocks, tiles_k),
            scratch_shapes=[pltpu.VMEM((tm, tn), jnp.float32)],
        ),
        input_output_aliases=input_output_aliases,
        compiler_params=pltpu.CompilerParams(dimension_semantics=(
            "parallel",
            "arbitrary",
            "arbitrary",
            "arbitrary",
        )),
        interpret=interpret,
        cost_estimate=cost_estimate,
    )

    out = call_gmm(
        group_metadata,
        group_offset,
        lhs,
        rhs,
        rhs_scale,
        rhs_bias,
        existing_out,
    )
    if existing_out is None and num_current_groups < num_total_groups:
        out = _zero_uninitialized_memory(
            out,
            start_group=group_offset[0],
            num_nonzero_groups=rhs.shape[0],
            group_metadata=group_metadata,
        )
    return out



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from dataclasses import InitVar, dataclass
from functools import partial
from typing import Optional

import jax
import jax.numpy as jnp
from flax import nnx
from flax.typing import Sharding
from jax.sharding import PartitionSpec
from jaxtyping import Float
from qwix._src.providers import ptq

from tpu_inference.kernels.fused_moe.v1.kernel import fused_ep_moe
from tpu_inference.layers.jax.base import create_param
from tpu_inference.layers.jax.layers import FlaxUtils
from tpu_inference.layers.jax.moe.dense_moe import (
    dense_moe_fwd, dense_moe_fwd_preapply_router_weights)
from tpu_inference.layers.jax.moe.sparse_moe import sparse_moe_distributed_fwd
from tpu_inference.layers.jax.moe.utils import MoEBackend
from tpu_inference.layers.vllm.fused_moe import fused_moe_func
from tpu_inference.models.jax.utils.qwix.qwix_utils import \
    manually_quantize_qwix_weight

modeling_flax_utils = FlaxUtils()


@dataclass(kw_only=True)
class CombineExperts(nnx.Module):
    """Combines expert outputs with router weights.

    Supports `TED,TE -> TD` when passed expert outputs, using float32
    accumulation for numerical stability, then casting back to the target
    dtype.
    """

    dtype: jnp.dtype

    def __call__(self, expert_outputs_TED: Float, weights_TE: Float) -> Float:
        with jax.named_scope("combine_experts"):
            output_TD = jnp.einsum(
                "TED,TE -> TD",
                expert_outputs_TED.astype(jnp.float32),
                weights_TE.astype(jnp.float32),
                precision="float32",
            )

        return output_TD.astype(self.dtype)


@dataclass(kw_only=True)
class Router(nnx.Module):
    """Router module for Mixture-of-Experts (MoE) layers.

    This module determines which experts each token should be routed to based on the input.

    Attributes:
    """
    dtype: jnp.dtype
    hidden_size: int
    num_experts: int
    num_experts_per_tok: int
    router_act: str
    rngs: InitVar[nnx.Rngs]
    activation_ffw_td: Sharding
    ed_sharding: Sharding
    random_init: bool = False
    moe_backend: MoEBackend = MoEBackend.DENSE_MAT

    def __call__(self, x_TD: Float):
        """Routes tokens to experts.

        Args:
            x_TD: Input array of shape (sequence_length, d_model).

        Returns:
            A tuple containing:
                - normalized_weights_TX: Normalized weights for selected experts, shape (sequence_length, num_experts_per_tok).
                - selected_experts_TX: Indices of selected experts, shape (sequence_length, num_experts_per_tok).
        """
        x_TD = jnp.asarray(x_TD, self.dtype)
        x_TD = nnx.with_sharding_constraint(x_TD, self.activation_ffw_td)
        router_act = modeling_flax_utils.ACT2FN[self.router_act]
        router_logits_TE = jnp.einsum('TD,DE -> TE', x_TD,
                                      self.kernel_DE.value)

        #TODO: Refactor the Router so that it will always only return router_logits_TE
        if self.moe_backend == MoEBackend.FUSED_MOE or self.moe_backend == MoEBackend.VLLM_MOE:
            return router_logits_TE
        else:
            weights_TX, selected_experts_TX = jax.lax.top_k(
                router_logits_TE, self.num_experts_per_tok)
            if self.router_act != "sigmoid":  # sigmoid does not accept axis argument.
                normalized_weights_TX = router_act(weights_TX.astype(
                    self.dtype),
                                                   axis=-1)
            else:
                normalized_weights_TX = router_act(
                    weights_TX.astype(self.dtype))
            return normalized_weights_TX, selected_experts_TX

    def __post_init__(self, rngs: nnx.Rngs):
        """Generates the router kernel (weights) for routing."""
        shape = (self.hidden_size, self.num_experts)
        self.kernel_DE = create_param(rngs,
                                      shape=shape,
                                      dtype=self.dtype,
                                      sharding=self.ed_sharding,
                                      random_init=self.random_init)


# --- Main Class for MoE ---
@dataclass(kw_only=True)
class MoE(nnx.Module):
    """Mixture-of-Experts (MoE) Routed MLP Layer.

    This module implements a MoE layer with a router and multiple expert MLPs.

    Attributes:
        router: The Router module.
    """
    dtype: jnp.dtype
    num_local_experts: int
    hidden_size: int
    intermediate_size_moe: int
    hidden_act: str
    rngs: InitVar[nnx.Rngs]
    router: nnx.Module
    mesh: jax.sharding.Mesh

    # --- Sharding Config ---
    activation_ffw_td: Sharding
    activation_ffw_ted: Sharding
    edf_sharding: Sharding
    efd_sharding: Sharding
    e2df_sharding: Sharding = ()

    # --- Flags & Configs ---
    apply_expert_weight_before_computation: bool
    random_init: bool = False
    moe_backend: MoEBackend = MoEBackend.DENSE_MAT

    # --- Sparse MoE Specific Attributes ---
    num_experts_per_tok: int = 1  # Required for Sparse, optional/derived for Dense
    tile_size: tuple[int, int, int] = (128, 128, 128)
    quantized_dtype: Optional[jnp.dtype] = None

    # --- MoE Kernel Specific Attributes ---
    renormalize: bool = True

    def __call__(self, x_TD: Float):
        """Performs the forward pass of the MoE layer.

        Args:
            x_TD: Input array of shape (sequence_length, d_model).

        Returns:
            Output array of shape (sequence_length, d_model) after passing through MoE.
        """
        x_TD = jnp.asarray(x_TD, self.dtype)
        x_TD = nnx.with_sharding_constraint(x_TD, self.activation_ffw_td)
        if self.moe_backend == MoEBackend.FUSED_MOE:
            router_logits_TE = self.router(x_TD)
            ep_axis_name = self.efd_sharding[0]
            output_TD = fused_ep_moe(
                mesh=self.mesh,
                tokens=x_TD,
                w1=self.kernel_gating_upproj_E2DF.value,
                w2=self.kernel_down_proj_EFD.value,
                gating_output=router_logits_TE,
                top_k=self.router.num_experts_per_tok,
                ep_axis_name=ep_axis_name,
                renormalize_topk_logits=self.renormalize,
                act_fn=self.hidden_act,
                **self.block_size,
            )
            return output_TD
        elif self.moe_backend == MoEBackend.VLLM_MOE:
            router_logits_TE = self.router(x_TD)
            output_TD = fused_moe_func(
                hidden_states=x_TD,
                w1=self.kernel_gating_upproj_EFD.value,
                w2=self.kernel_down_proj_EDF.value,
                w1_bias=self.w1_bias,
                w2_bias=self.w2_bias,
                w1_scale=self.w1_scale,
                w2_scale=self.w2_scale,
                gating_output=router_logits_TE,
                topk=self.router.num_experts_per_tok,
                renormalize=self.renormalize,
                mesh=self.mesh,
                use_ep=self.num_expert_parallelism > 1,
                activation=self.hidden_act,
            )
            return output_TD
        else:
            weights_TX, indices_TX = self.router(x_TD)

            if self.moe_backend == MoEBackend.MEGABLX_GMM or self.moe_backend == MoEBackend.RAGGED_DOT:
                in_specs = (
                    PartitionSpec(),  # replicated MoE instance
                    PartitionSpec(*self.activation_ffw_td),  # Sharded x_TD
                    PartitionSpec(),  # Replicated router_weights_TX
                    PartitionSpec(),  # Replicated selected_experts_TX
                    PartitionSpec(*self.edf_sharding),  # Sharded gating kernel
                    PartitionSpec(
                        *self.edf_sharding),  # Sharded up-projection kernel
                    PartitionSpec(
                        *self.efd_sharding),  # Sharded down-projection kernel
                )
                out_specs = PartitionSpec(*self.activation_ffw_td)

                mapped_moe_fwd = partial(
                    jax.experimental.shard_map.shard_map,
                    mesh=self.mesh,
                    in_specs=in_specs,
                    out_specs=out_specs,
                    check_rep=False)(sparse_moe_distributed_fwd)

                kernel_gating_EDF = self._process_weight_for_qwix(
                    self.kernel_gating_EDF,
                    channelwise_axes=[0, 2],
                    tiled_axes={})
                kernel_up_proj_EDF = self._process_weight_for_qwix(
                    self.kernel_up_proj_EDF,
                    channelwise_axes=[0, 2],
                    tiled_axes={})
                kernel_down_proj_EFD = self._process_weight_for_qwix(
                    self.kernel_down_proj_EFD,
                    channelwise_axes=[0, 2],
                    tiled_axes={})

                return mapped_moe_fwd(self, x_TD, weights_TX, indices_TX,
                                      kernel_gating_EDF, kernel_up_proj_EDF,
                                      kernel_down_proj_EFD)

            # Dense Matmul
            elif self.moe_backend == MoEBackend.DENSE_MAT:
                one_hot_indices_TXE = jax.nn.one_hot(
                    indices_TX,
                    num_classes=self.num_local_experts,
                    dtype=self.dtype)
                full_weights_TE = jnp.sum(one_hot_indices_TXE *
                                          weights_TX[..., None],
                                          axis=1)
                # Some models use the routing scores to weight the data instead of
                # weighting the expert outputs.
                if self.apply_expert_weight_before_computation:
                    with jax.named_scope("pre_computing_weight"):
                        return dense_moe_fwd_preapply_router_weights(
                            self, x_TD, full_weights_TE)
                else:
                    return dense_moe_fwd(self, x_TD, full_weights_TE)

    def __post_init__(self, rngs: nnx.Rngs):
        """Generates the kernels (weights) for the router and experts (gating, up-projection, and down-projection layers)."""
        E = self.num_local_experts
        D = self.hidden_size
        F = self.intermediate_size_moe

        if self.moe_backend == MoEBackend.FUSED_MOE:
            if self.edf_sharding:
                self.e2df_sharding = (self.edf_sharding[0], None,
                                      self.edf_sharding[1],
                                      self.edf_sharding[2])
            self.kernel_gating_upproj_E2DF = create_param(
                rngs,
                shape=(E, 2, D, F),
                dtype=self.dtype,
                sharding=self.e2df_sharding,
                random_init=self.random_init)
            self.kernel_down_proj_EFD = create_param(
                rngs,
                shape=(E, F, D),
                dtype=self.dtype,
                sharding=self.efd_sharding,
                random_init=self.random_init)
            self.block_size = {
                "bt": 32,
                "bf": 512,
                "bd1": 512,
                "bd2": 512,
                "btc": 64,
                "bfc": 256,
                "bd1c": 256,
                "bd2c": 256,
            }
        elif self.moe_backend == MoEBackend.VLLM_MOE:
            self.kernel_gating_upproj_EFD = create_param(
                rngs,
                shape=(E, 2 * F, D),
                dtype=self.dtype,
                sharding=self.efd_sharding,
                random_init=self.random_init)
            self.kernel_down_proj_EDF = create_param(
                rngs,
                shape=(E, D, F),
                dtype=self.dtype,
                sharding=self.edf_sharding,
                random_init=self.random_init)
        else:
            self.kernel_gating_EDF = create_param(rngs,
                                                  shape=(E, D, F),
                                                  dtype=self.dtype,
                                                  sharding=self.edf_sharding,
                                                  random_init=self.random_init)
            self.kernel_up_proj_EDF = create_param(
                rngs,
                shape=(E, D, F),
                dtype=self.dtype,
                sharding=self.edf_sharding,
                random_init=self.random_init)
            self.kernel_down_proj_EFD = create_param(
                rngs,
                shape=(E, F, D),
                dtype=self.dtype,
                sharding=self.efd_sharding,
                random_init=self.random_init)

        # Default MoE has no bias vectors
        self.w1_bias, self.w2_bias = (None, None)

        # TODO: Add quantization scale params for VLLM MoE kernel
        self.w1_scale, self.w2_scale = (None, None)

        self.expert_axis_name = self.edf_sharding[0]
        if self.expert_axis_name is None:
            self.num_expert_parallelism = 1
        else:
            if isinstance(self.expert_axis_name, str):
                self.num_expert_parallelism = self.mesh.shape[
                    self.expert_axis_name]
            else:
                self.num_expert_parallelism = math.prod(
                    self.mesh.shape[axis] for axis in self.expert_axis_name)

        # Derive if data is sharded by expert
        self.data_axis_name = self.activation_ffw_td[0]
        self.is_batch_sharded_by_expert = (
            self.expert_axis_name is not None) and (self.expert_axis_name
                                                    == self.data_axis_name)

    def _process_weight_for_qwix(self,
                                 weight_param,
                                 channelwise_axes=[],
                                 tiled_axes={}):
        """
        Extracts weight value, applies quantization if needed,
        and returns the underlying array.
        """
        weight = weight_param.value

        if self.quantized_dtype:
            if not isinstance(weight, ptq.WithAux):
                weight = manually_quantize_qwix_weight(weight,
                                                       self.quantized_dtype,
                                                       channelwise_axes,
                                                       tiled_axes, "absmax")
            return weight.array

        return weight



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import jax
import numpy as np
import pytest
from vllm.config import (CacheConfig, ModelConfig, ParallelConfig,
                         SchedulerConfig, SpeculativeConfig, VllmConfig)
from vllm.sampling_params import SamplingType
from vllm.v1.outputs import DraftTokenIds

from tpu_inference.runner.input_batch import CachedRequestState, InputBatch
from tpu_inference.runner.speculative_decoding_manager import \
    SpecDecodeMetadata
from tpu_inference.runner.tpu_runner import TPUModelRunner
from tpu_inference.spec_decode.jax.eagle3 import Eagle3Proposer


class TestSpeculativeDecodingManager:

    def setup_method(self):
        # Mock JAX dependencies
        self.mock_devices = [MagicMock(coords=i) for i in range(1)]
        device_array = np.array(jax.devices()[:1]).reshape(1, 1, 1, 1)
        self.mock_mesh = jax.make_mesh(device_array.shape,
                                       ('data', 'attn_dp', 'expert', 'model'))
        self.mock_rng_key = MagicMock()

        with patch('jax.devices', return_value=self.mock_devices), \
             patch('jax.make_mesh', return_value=self.mock_mesh), \
             patch('jax.random.key', return_value=self.mock_rng_key), \
             patch('tpu_inference.runner.tpu_runner.get_model', return_value=MagicMock()), \
             patch('tpu_inference.runner.tpu_runner.make_optimized_mesh', return_value=self.mock_mesh):

            model_config = ModelConfig(tokenizer_mode="auto",
                                       trust_remote_code=False,
                                       seed=0,
                                       dtype='bfloat16')
            cache_config = CacheConfig(
                block_size=16,
                gpu_memory_utilization=0.9,
                swap_space=4,
                cache_dtype="auto",
            )
            scheduler_config = SchedulerConfig(max_num_seqs=16,
                                               max_model_len=1024,
                                               is_encoder_decoder=False)
            parallel_config = ParallelConfig(
                pipeline_parallel_size=1,
                tensor_parallel_size=1,
                worker_use_ray=False,
            )
            speculative_config = SpeculativeConfig(
                model='ngram',
                num_speculative_tokens=5,
                prompt_lookup_max=4,
            )
            vllm_config = VllmConfig(
                model_config=model_config,
                cache_config=cache_config,
                scheduler_config=scheduler_config,
                parallel_config=parallel_config,
                speculative_config=speculative_config,
                observability_config={},
                additional_config={},
            )

            self.runner = TPUModelRunner(vllm_config,
                                         devices=self.mock_devices)

    def test_propose_draft_token_ids_dispatches_to_eagle(self):
        """Tests that propose_draft_token_ids calls the correct eagle method."""
        # 1. ===== Setup =====
        # Set the drafter to be an Eagle3Proposer
        self.runner.drafter = MagicMock(spec=Eagle3Proposer)
        self.runner.speculative_config.method = "eagle3"

        # Mock the eagle-specific proposal method
        with patch.object(self.runner.speculative_decoding_manager,
                          'propose_eagle3_draft_token_ids',
                          return_value=[[10, 11]]) as mock_propose_eagle:

            # 2. ===== Act =====
            self.runner.speculative_decoding_manager.propose_draft_token_ids(
                sampled_token_ids=[[1]],
                aux_hidden_states=None,
                attn_metadata=MagicMock(),
                spec_decode_metadata=None,
            )

            # 3. ===== Assert =====
            mock_propose_eagle.assert_called_once()
            assert self.runner.speculative_decoding_manager._draft_token_ids == [
                [10, 11]
            ]

    def test_propose_draft_token_ids_wrong_drafter_type(self):
        """Tests that an assertion is raised if the drafter is not an NgramProposer."""
        # The default drafter is NgramProposer, so we replace it with a generic mock
        self.runner.drafter = MagicMock()
        self.runner.speculative_config.method = "ngram"
        with pytest.raises(AssertionError):
            self.runner.speculative_decoding_manager.propose_draft_token_ids(
                [[1]], None, MagicMock(), None)

    def test_take_draft_token_ids(self):
        """Tests the take_draft_token_ids method for speculative decoding."""
        # Case 1: No draft tokens are available.
        self.runner.speculative_decoding_manager._draft_token_ids = None
        result = self.runner.take_draft_token_ids()
        assert result is None

        # Case 2: Draft tokens are available.
        mock_req_ids = ["req-1", "req-2"]
        mock_draft_ids = [[10, 11], [20, 21, 22]]

        # Re-initialize input_batch for a clean state for this specific test
        self.runner.input_batch = InputBatch(
            max_num_reqs=self.runner.max_num_reqs,
            max_model_len=self.runner.max_model_len,
            max_num_batched_tokens=self.runner.max_num_tokens,
            pin_memory=False,
            vocab_size=self.runner.vocab_size,
            block_sizes=[self.runner.block_size],
            is_spec_decode=True,
        )

        # Add some requests to populate `input_batch.req_ids`
        mock_sampling_params = MagicMock()
        mock_sampling_params.sampling_type = SamplingType.GREEDY
        mock_sampling_params.top_k = -1
        mock_sampling_params.top_p = 1.0
        mock_sampling_params.temperature = 0.0
        mock_sampling_params.min_tokens = 0
        mock_sampling_params.logprobs = None
        mock_sampling_params.logit_bias = None
        mock_sampling_params.allowed_token_ids = set()
        mock_sampling_params.bad_words_token_ids = None
        mock_sampling_params.all_stop_token_ids = set()

        req1 = CachedRequestState(req_id="req-1",
                                  prompt_token_ids=[1],
                                  output_token_ids=[],
                                  sampling_params=mock_sampling_params,
                                  block_ids=([1], ),
                                  num_computed_tokens=1,
                                  lora_request=None,
                                  mm_features=[],
                                  pooling_params=None,
                                  generator=None)
        req2 = CachedRequestState(req_id="req-2",
                                  prompt_token_ids=[2],
                                  output_token_ids=[],
                                  sampling_params=mock_sampling_params,
                                  block_ids=([2], ),
                                  num_computed_tokens=1,
                                  lora_request=None,
                                  mm_features=[],
                                  pooling_params=None,
                                  generator=None)
        self.runner.input_batch.add_request(req1)
        self.runner.input_batch.add_request(req2)

        # Set the draft tokens to be taken
        self.runner.speculative_decoding_manager._draft_token_ids = mock_draft_ids

        # Call the method to be tested
        result = self.runner.take_draft_token_ids()

        # Assertions for the returned object
        assert result is not None
        assert isinstance(result, DraftTokenIds)
        assert result.req_ids == mock_req_ids
        assert result.draft_token_ids == mock_draft_ids

        # Assert that the internal state is reset
        assert self.runner.speculative_decoding_manager._draft_token_ids is None

        # Case 3: Call again after taking, should return None
        result_after = self.runner.take_draft_token_ids()
        assert result_after is None

    def _setup_spec_decode_metadata_test(self):
        """Helper method to set up common test infrastructure for spec decode metadata tests."""
        # Mock runner attributes needed by the function
        self.runner.arange_cpu = np.arange(1024, dtype=np.int64)
        # Make input_ids_cpu a sequence of numbers for easy verification
        self.runner.input_ids_cpu = np.arange(1024, dtype=np.int32) * 10
        self.runner.num_tokens_paddings = [16, 32, 64, 128, 256, 512, 1024]

        # Mock the device_array function to just return the numpy arrays
        def mock_device_array(mesh, *args, **kwargs):
            # Skip mesh parameter and return the actual arrays
            if len(args) == 1 and isinstance(args[0], tuple):
                return args[0]
            return args

        self.mock_device_array = mock_device_array

    @pytest.mark.parametrize(
        "num_draft_tokens,cu_num_scheduled_tokens,padded_num_reqs,expected_logits_indices,expected_bonus_logits_indices,expected_target_logits_indices,expected_draft_token_ids",
        [
            (
                # Normal case
                [3, 0, 2, 0, 1],
                [4, 104, 107, 207, 209],
                8,
                [0, 1, 2, 3, 103, 104, 105, 106, 206, 207, 208],
                [3, 4, 7, 8, 10, 0, 0, 0],
                [0, 1, 2, 5, 6, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [10, 20, 30, 1050, 1060, 2080, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
            (
                # High speculative tokens case
                [5, 3, 4, 2, 1],
                [6, 10, 18, 22, 26],
                8,
                [
                    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 13, 14, 15, 16, 17, 19, 20,
                    21, 24, 25
                ],
                [5, 9, 14, 17, 19, 0, 0, 0],
                [
                    0, 1, 2, 3, 4, 6, 7, 8, 10, 11, 12, 13, 15, 16, 18, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
                ],
                [
                    10, 20, 30, 40, 50, 70, 80, 90, 140, 150, 160, 170, 200,
                    210, 250
                ]),
        ])
    def test_get_spec_decode_metadata_parametrized(
            self, num_draft_tokens, cu_num_scheduled_tokens, padded_num_reqs,
            expected_logits_indices, expected_bonus_logits_indices,
            expected_target_logits_indices, expected_draft_token_ids):
        """Comprehensive parametrized test for _get_spec_decode_metadata function."""
        # Setup
        self._setup_spec_decode_metadata_test()

        # Convert Python lists to numpy arrays for function input
        num_draft_tokens_np = np.array(num_draft_tokens, dtype=np.int32)
        cu_num_scheduled_tokens_np = np.array(cu_num_scheduled_tokens,
                                              dtype=np.int32)

        # Act
        with patch(
                "tpu_inference.runner.speculative_decoding_manager.device_array",
                side_effect=self.mock_device_array):
            metadata = self.runner.speculative_decoding_manager.get_spec_decode_metadata(
                num_draft_tokens_np,
                cu_num_scheduled_tokens_np,
                padded_num_reqs=padded_num_reqs)

        # Assert basic properties
        assert isinstance(metadata, SpecDecodeMetadata)

        # Determine padding length based on expected_logits_indices length
        if len(expected_logits_indices) <= 16:
            padded_len = 16
        else:
            padded_len = 32

        # final_logits_indices - pad to bucket size and compare as Python lists
        expected_padded_logits_indices = expected_logits_indices + [0] * (
            padded_len - len(expected_logits_indices))
        assert np.asarray(metadata.final_logits_indices).tolist(
        ) == expected_padded_logits_indices

        # bonus_logits_indices - compare as Python lists
        assert np.asarray(metadata.bonus_logits_indices).tolist(
        ) == expected_bonus_logits_indices

        # target_logits_indices - pad to same length as final_logits_indices and compare as Python lists
        expected_padded_target_logits_indices = expected_target_logits_indices + [
            0
        ] * (padded_len - len(expected_target_logits_indices))
        assert np.asarray(metadata.target_logits_indices).tolist(
        ) == expected_padded_target_logits_indices

        # draft_token_ids - pad the expected values to the correct length and compare as Python lists
        expected_padded_draft_token_ids = expected_draft_token_ids + [0] * (
            padded_len - len(expected_draft_token_ids))
        assert np.asarray(metadata.draft_token_ids).tolist(
        ) == expected_padded_draft_token_ids

        # draft_lengths - pad and compare as Python lists
        expected_padded_num_draft_tokens = num_draft_tokens + [0] * (
            padded_num_reqs - len(num_draft_tokens))
        assert np.asarray(metadata.draft_lengths).tolist(
        ) == expected_padded_num_draft_tokens

    @pytest.mark.parametrize("spec_decode_metadata_is_none", [True, False])
    def test_propose_eagle3_draft_token_ids(self,
                                            spec_decode_metadata_is_none):
        """Tests the logic for proposing Eagle3 draft tokens."""
        # 1. ===== Setup =====
        self.runner.drafter = MagicMock(spec=Eagle3Proposer)
        self.runner.speculative_config.method = "eagle3"

        # Mock TPUModelRunner attributes
        self.runner.input_batch = MagicMock()
        self.runner.input_batch.req_ids = ["req-1", "req-2"]
        self.runner.requests = {
            "req-1": MagicMock(),
            "req-2": MagicMock(),
        }
        self.runner.mesh = self.mock_mesh
        self.runner.kv_caches = MagicMock()

        # Mock drafter methods
        mock_attn_metadata = MagicMock()
        mock_target_token_ids = MagicMock()
        mock_last_token_indices = MagicMock()
        mock_target_hidden_states = MagicMock()
        self.runner.drafter.prepare_inputs.return_value = (
            mock_target_hidden_states,
            mock_target_token_ids,
            mock_last_token_indices,
            mock_attn_metadata,
        )
        mock_draft_token_ids = [[10, 11], [20, 21]]
        self.runner.drafter.propose.return_value = (
            self.runner.kv_caches,
            mock_draft_token_ids,
        )

        # Inputs
        sampled_token_ids = [[1], [2]]
        aux_hidden_states = MagicMock()
        attn_metadata = MagicMock()
        attn_metadata.seq_lens.shape = [2]
        if spec_decode_metadata_is_none:
            spec_decode_metadata = None
        else:
            spec_decode_metadata = MagicMock(spec=SpecDecodeMetadata)
            spec_decode_metadata.draft_lengths_cpu = np.array([2, 3])
        scheduler_output = MagicMock()
        input_ids = MagicMock()

        # 2. ===== Act =====
        with patch(
                "tpu_inference.runner.speculative_decoding_manager.device_array",
                side_effect=lambda mesh, x: x):
            result = self.runner.speculative_decoding_manager.propose_eagle3_draft_token_ids(
                sampled_token_ids,
                aux_hidden_states,
                attn_metadata,
                spec_decode_metadata,
                scheduler_output,
                input_ids,
            )

        # 3. ===== Assert =====
        assert result == [[10, 11], [20, 21]]
        self.runner.drafter.prepare_inputs.assert_called_once()
        self.runner.drafter.propose.assert_called_once()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import pytest
from vllm.config import ModelConfig
from vllm.lora.request import LoRARequest
from vllm.v1.kv_cache_interface import KVCacheConfig
from vllm.v1.outputs import DraftTokenIds

# The class we are testing
from tpu_inference.worker.tpu_worker import TPUWorker


@pytest.fixture
def mock_vllm_config():
    """
    Provides a mock VllmConfig object for tests.
    This version builds the mock explicitly to avoid spec-related AttributeErrors.
    """
    # Create mocks for the nested config objects first
    mock_cache_conf = MagicMock()
    mock_cache_conf.gpu_memory_utilization = 0.9
    mock_cache_conf.num_gpu_blocks = 0
    mock_cache_conf.num_cpu_blocks = 0

    mock_parallel_conf = MagicMock()
    mock_parallel_conf.tensor_parallel_size = 2
    mock_parallel_conf.data_parallel_size = 1
    mock_parallel_conf.pipeline_parallel_size = 1
    mock_parallel_conf.nnodes = 1
    mock_parallel_conf.nnodes_within_dp = 1

    mock_additional_config = {}

    # Create the main config mock and attach the others without a top-level spec
    config = MagicMock()
    config.model_config = ModelConfig(model="Qwen/Qwen3-0.6B")
    config.cache_config = mock_cache_conf
    config.parallel_config = mock_parallel_conf
    config.additional_config = mock_additional_config

    config.sharding_config = MagicMock()
    config.sharding_config.total_devices = 2

    return config


@pytest.fixture
def mock_get_pp_group():
    with patch("tpu_inference.distributed.jax_parallel_state.get_pp_group",
               return_value=MagicMock(is_first_rank=True,
                                      is_last_rank=True,
                                      rank_in_group=0,
                                      world_size=1)):
        yield


class TestTPUWorker:
    """Test suite for the TPUWorker class."""

    #
    # --- Initialization Tests ---
    #

    def test_init_success(self, mock_vllm_config):
        """Tests successful initialization of TPUWorker."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test_method",
                           is_driver_worker=True,
                           devices=['tpu:0'])
        assert worker.vllm_config == mock_vllm_config
        assert worker.rank == 0
        assert worker.local_rank == 0
        assert worker.is_driver_worker
        assert worker.profile_dir is None
        assert worker.devices == ['tpu:0']

    @patch('tpu_inference.worker.tpu_worker.vllm_envs')
    def test_init_with_profiler_on_rank_zero(self, mock_envs,
                                             mock_vllm_config):
        """Tests that the profiler directory is set correctly on rank 0."""
        mock_envs.VLLM_TORCH_PROFILER_DIR = "/tmp/profiles"
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test_method")
        assert worker.profile_dir == "/tmp/profiles"

    @patch('tpu_inference.worker.tpu_worker.vllm_envs')
    def test_init_with_profiler_on_other_ranks(self, mock_envs,
                                               mock_vllm_config):
        """Tests that the profiler directory is NOT set on non-rank 0 workers."""
        mock_envs.VLLM_TORCH_PROFILER_DIR = "/tmp/profiles"
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=1,
                           rank=1,
                           distributed_init_method="test_method")
        assert worker.profile_dir is None

    #
    # --- Device and Cache Initialization Tests ---
    #

    def test_initialize_cache(self, mock_vllm_config):
        """Tests setting the number of GPU and CPU cache blocks."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test_method")
        worker.initialize_cache(num_gpu_blocks=2048, num_cpu_blocks=1024)
        assert worker.cache_config.num_gpu_blocks == 2048
        assert worker.cache_config.num_cpu_blocks == 1024

    @patch('tpu_inference.worker.tpu_worker.TPUModelRunner')
    @patch('tpu_inference.worker.tpu_worker.utils')
    @patch('tpu_inference.worker.tpu_worker.jax')
    @patch('tpu_inference.worker.tpu_worker.ensure_kv_transfer_initialized')
    def test_init_device_with_provided_devices(
            self, mock_ensure_kv_transfer_initialized, mock_jax, mock_utils,
            mock_runner_cls, mock_vllm_config, mock_get_pp_group):
        """Tests init_device when devices are provided during construction."""
        mock_devices = ['tpu:0', 'tpu:1']
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test_method",
                           devices=mock_devices)

        worker.init_device()

        expected_rank = 0
        expected_is_first_rank = True
        expected_is_last_rank = True
        mock_runner_cls.assert_called_once_with(mock_vllm_config, mock_devices,
                                                expected_rank,
                                                expected_is_first_rank,
                                                expected_is_last_rank)
        assert isinstance(worker.model_runner, MagicMock)

    @patch('tpu_inference.worker.tpu_worker.TPUModelRunner')
    @patch('tpu_inference.worker.tpu_worker.utils')
    @patch('tpu_inference.worker.tpu_worker.jax')
    @patch('tpu_inference.worker.tpu_worker.ensure_kv_transfer_initialized')
    def test_init_device_autodetects_devices(
            self, mock_ensure_kv_transfer_initialized, mock_jax, mock_utils,
            mock_runner_cls, mock_vllm_config):
        """Tests init_device when devices are auto-detected via JAX."""
        worker = TPUWorker(
            vllm_config=mock_vllm_config,
            local_rank=0,
            rank=0,
            distributed_init_method="test_method",
            devices=[]  # No devices provided, should trigger auto-detection
        )
        mock_jax.device_count.return_value = 4
        mock_jax.devices.return_value = ['tpu:0', 'tpu:1', 'tpu:2', 'tpu:3']

        worker.init_device()

        expected_devices = ['tpu:0', 'tpu:1']  # Sliced by tensor_parallel_size
        assert worker.devices == expected_devices
        expected_rank = 0
        expected_is_first_rank = True
        expected_is_last_rank = True
        mock_runner_cls.assert_called_once_with(mock_vllm_config,
                                                expected_devices,
                                                expected_rank,
                                                expected_is_first_rank,
                                                expected_is_last_rank)

    @patch('tpu_inference.worker.tpu_worker.utils')
    def test_determine_available_memory(self, mock_utils, mock_vllm_config):
        """Tests the available HBM memory calculation."""
        # Setup mock return for hbm_usage_bytes: [(used_bytes, limit_bytes), ...]
        mock_utils.hbm_usage_bytes.return_value = [
            (100 * 1024**3, 1000 * 1024**3), (200 * 1024**3, 1000 * 1024**3)
        ]
        mock_devices = ['tpu:0', 'tpu:1']
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test_method",
                           devices=mock_devices)

        available_mem = worker.determine_available_memory()

        mock_utils.hbm_usage_bytes.assert_called_once_with(mock_devices)
        # Total limit: 1000 + 1000 = 2000 GiB
        # Total cap: 2000 * 0.9 = 1800 GiB
        # Total used: 100 + 200 = 300 GiB
        # Total free = 1800 - 300 = 1500 GiB
        expected_mem = 1500 * 1024**3
        assert available_mem == expected_mem

    #
    # --- Core Logic Tests ---
    #

    @patch('tpu_inference.worker.tpu_worker.TPUModelRunner')
    def test_execute_model(self, mock_runner_cls, mock_vllm_config):
        """Tests that the driver worker executes the model and returns the concrete vLLM output."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test",
                           is_driver_worker=True)
        worker.model_runner = mock_runner_cls.return_value  # Assign mocked runner instance
        mock_scheduler_input = MagicMock()

        # The model runner returns a concrete vllm output
        mock_model_output = "concrete_model_output"
        worker.model_runner.execute_model.return_value = mock_model_output

        result = worker.execute_model(mock_scheduler_input)

        # Assert the runner was called with the scheduler output directly
        worker.model_runner.execute_model.assert_called_once_with(
            mock_scheduler_input, None)
        # Assert the final result is the concrete model output
        assert result == mock_model_output

    @patch('tpu_inference.worker.tpu_worker.TPUModelRunner')
    def test_execute_model_non_driver_returns_none(self, mock_runner_cls,
                                                   mock_vllm_config):
        """Tests that a non-driver worker executes the model but returns None."""
        worker = TPUWorker(
            vllm_config=mock_vllm_config,
            local_rank=0,
            rank=0,
            distributed_init_method="test",
            is_driver_worker=False  # Not a driver
        )
        worker.model_runner = mock_runner_cls.return_value
        mock_scheduler_input = MagicMock()

        result = worker.execute_model(mock_scheduler_input)

        assert result is None

    def test_take_draft_token_ids(self, mock_vllm_config):
        """Tests that take_draft_token_ids correctly passes through from the runner."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test")
        worker.model_runner = MagicMock()

        # Case 1: Runner returns a DraftTokenIds object
        mock_draft_tokens = DraftTokenIds(req_ids=["req1"],
                                          draft_token_ids=[[1, 2]])
        worker.model_runner.take_draft_token_ids.return_value = mock_draft_tokens

        result = worker.take_draft_token_ids()
        worker.model_runner.take_draft_token_ids.assert_called_once()
        assert result == mock_draft_tokens

    def test_add_lora_not_implemented(self, mock_vllm_config):
        """Tests that add_lora raises NotImplementedError."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test")
        mock_lora_request = MagicMock()

        with pytest.raises(
                NotImplementedError,
                match="LoRA is not supported by the JAX worker yet."):
            worker.add_lora(mock_lora_request)

    def test_add_lora_not_implemented_lora_request(self, mock_vllm_config):
        """Tests that add_lora raises NotImplementedError."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test")
        mock_lora_request = MagicMock(spec=LoRARequest)

        with pytest.raises(
                NotImplementedError,
                match="LoRA is not supported by the JAX worker yet."):
            worker.add_lora(mock_lora_request)

    #
    # --- Profiling and Health Check Tests ---
    #

    @patch('tpu_inference.worker.tpu_worker.jax')
    @patch.dict('os.environ', {"PYTHON_TRACER_LEVEL": "1"}, clear=True)
    def test_profile_start(self, mock_jax, mock_vllm_config):
        """Tests starting the JAX profiler."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test")
        worker.profile_dir = "/tmp/profile_dir"

        worker.profile(is_start=True)

        mock_jax.profiler.ProfileOptions.assert_called_once()
        mock_jax.profiler.start_trace.assert_called_once()
        args, kwargs = mock_jax.profiler.start_trace.call_args
        assert args[0] == "/tmp/profile_dir"
        # Verify options from env var were used
        assert kwargs['profiler_options'].python_tracer_level == 1

    @patch('tpu_inference.worker.tpu_worker.jax')
    def test_profile_stop(self, mock_jax, mock_vllm_config):
        """Tests stopping the JAX profiler."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test")
        worker.profile(is_start=False)
        mock_jax.profiler.stop_trace.assert_called_once()

    def test_check_health(self, mock_vllm_config):
        """Tests that check_health runs without error."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test")
        try:
            worker.check_health()
        except Exception as e:
            pytest.fail(
                f"TPUWorker.check_health() raised an unexpected exception: {e}"
            )

    #
    # --- Pass-through Method Tests ---
    #

    @pytest.mark.parametrize(
        "worker_method_name, runner_method_name, method_args", [
            ("load_model", "load_model", []),
            ("get_model", "get_model", []),
            ("get_kv_cache_spec", "get_kv_cache_spec", []),
        ])
    def test_runner_passthrough_methods(self, worker_method_name,
                                        runner_method_name, method_args,
                                        mock_vllm_config):
        """Tests methods that are simple pass-throughs to the TPUModelRunner."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test")
        worker.model_runner = MagicMock()

        # Call the worker method and assert the underlying runner method was called
        getattr(worker, worker_method_name)(*method_args)
        mock_runner_method = getattr(worker.model_runner, runner_method_name)
        mock_runner_method.assert_called_once_with(*method_args)

    def test_initialize_from_config(self, mock_vllm_config):
        """Tests the special case pass-through for initialize_from_config."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test")
        worker.model_runner = MagicMock()
        worker.topology_order_id = 0
        mock_input_config = MagicMock()

        worker.initialize_from_config(mock_input_config)

        worker.model_runner.initialize_kv_cache.assert_called_once_with(
            mock_input_config, 0)

    def test_initialize_from_config_kv_cache_config(self, mock_vllm_config):
        """Tests the special case pass-through for initialize_from_config."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test")
        worker.model_runner = MagicMock()
        worker.topology_order_id = 0
        mock_input_config = MagicMock(spec=KVCacheConfig)

        worker.initialize_from_config(mock_input_config)

        worker.model_runner.initialize_kv_cache.assert_called_once_with(
            mock_input_config, 0)

    def test_compile_or_warm_up_model(self, mock_vllm_config):
        """Tests the special case pass-through for model compilation/warmup."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test")
        worker.model_runner = MagicMock()

        worker.compile_or_warm_up_model()

        # This method calls two different runner methods
        worker.model_runner.capture_model.assert_called_once()
        worker.model_runner._init_random.assert_called_once()

    def test_get_supported_tasks(self, mock_vllm_config):
        """Test get_supported_tasks passthrough to model runner."""
        worker = TPUWorker(vllm_config=mock_vllm_config,
                           local_rank=0,
                           rank=0,
                           distributed_init_method="test")
        worker.model_runner = MagicMock()
        worker.model_runner.get_supported_tasks.return_value = ("generate", )

        _ = worker.get_supported_tasks()

        worker.model_runner.get_supported_tasks.assert_called_once()



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Sequence

import jax
from jax import numpy as jnp

from tpu_inference.layers.common.quantization.configs import QuantLinearConfig
from tpu_inference.layers.common.utils import \
    slice_sharded_tensor_for_concatenation


class UnquantizedLinearMethod:
    """Implements the forward method for unquantized linear layers.
    
    This class will be shared in both vLLM and jax path.
    """

    def __init__(self, linear_config: QuantLinearConfig):
        self.linear_config = linear_config

    def _apply_fused(self,
                     x_jax: jax.Array,
                     weight_jax: jax.Array,
                     bias_jax: Optional[jax.Array] = None,
                     einsum_str: str = "mn,pn->mp") -> jax.Array:
        outs = jnp.einsum(einsum_str, x_jax, weight_jax)
        if bias_jax is not None:
            outs += bias_jax

        outs = slice_sharded_tensor_for_concatenation(
            outs, self.linear_config.output_sizes, self.linear_config.n_shards)
        out = jnp.concatenate(outs, axis=-1)
        return out

    def _apply_split(self,
                     x_jax: jax.Array,
                     weights: Sequence[jax.Array],
                     bias_jax: Optional[jax.Array] = None) -> jax.Array:
        outs = []
        for i, weight_jax in enumerate(weights):
            out = jnp.einsum("mn,pn->mp", x_jax, weight_jax)
            if bias_jax is not None:
                out += bias_jax[i]

            outs.append(out)
        out = jnp.concatenate(outs, axis=-1)
        return out



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

import jax
import jax.numpy as jnp
import numpy as np
import pytest
from jax.sharding import Mesh, NamedSharding, PartitionSpec

from tpu_inference.layers.jax.sample.sampling_metadata import (
    DEFAULT_SAMPLING_PARAMS, TPUSupportedSamplingMetadata)

## Mocks and Fixtures


@dataclass
class MockInputBatch:
    """A mock of the InputBatch class, using NumPy arrays for CPU tensors."""

    all_greedy: bool
    num_reqs: int = 0
    temperature_cpu: np.ndarray = None
    top_k_cpu: np.ndarray = None
    top_p_cpu: np.ndarray = None
    max_num_logprobs: int = None


@pytest.fixture(scope="module")
def mesh() -> Mesh:
    """Creates a 1D JAX mesh for testing on available devices."""
    if not jax.devices():
        pytest.skip("No JAX devices available for testing.")
    return Mesh(np.array(jax.devices()), axis_names=("data", ))


## Test Cases


def test_from_input_batch_all_greedy(mesh: Mesh):
    """
    Tests TPUSupportedSamplingMetadata.from_input_batch when **all_greedy is True**.

    It should return an object with `do_sampling=False` and `None` for the tensors.
    """
    mock_batch = MockInputBatch(all_greedy=True)
    padded_num_reqs = 4

    metadata = TPUSupportedSamplingMetadata.from_input_batch(
        mesh=mesh, input_batch=mock_batch, padded_num_reqs=padded_num_reqs)

    assert not metadata.do_sampling, "do_sampling should be False for greedy requests"
    assert metadata.temperature is None
    assert metadata.top_k is None
    assert metadata.top_p is None


def test_from_input_batch_with_sampling_and_padding(mesh: Mesh):
    """
    Tests TPUSupportedSamplingMetadata.from_input_batch with sampling enabled,
    requiring the tensors to be **padded** to the correct shape.
    """
    num_reqs = 2
    padded_num_reqs = 4

    # Input tensors must be large enough to hold the padded values.
    temp_tensor = np.array([0.7, 0.8, 0.0, 0.0], dtype=np.float32)
    top_k_tensor = np.array([10, 20, 0, 0], dtype=np.int32)
    top_p_tensor = np.array([0.9, 0.95, 0.0, 0.0], dtype=np.float32)

    mock_batch = MockInputBatch(
        all_greedy=False,
        num_reqs=num_reqs,
        temperature_cpu=temp_tensor,
        top_k_cpu=top_k_tensor,
        top_p_cpu=top_p_tensor,
    )

    metadata = TPUSupportedSamplingMetadata.from_input_batch(
        mesh=mesh, input_batch=mock_batch, padded_num_reqs=padded_num_reqs)

    # 1. Check metadata flags and types
    assert metadata.do_sampling, "do_sampling should be True"
    assert isinstance(metadata.temperature, jnp.ndarray)
    assert isinstance(metadata.top_k, jnp.ndarray)
    assert isinstance(metadata.top_p, jnp.ndarray)

    # 2. Check shapes
    assert metadata.temperature.shape == (padded_num_reqs, )
    assert metadata.top_k.shape == (padded_num_reqs, )
    assert metadata.top_p.shape == (padded_num_reqs, )

    # 3. Check sharding (should be fully replicated)
    expected_sharding = NamedSharding(mesh, PartitionSpec(None))
    assert metadata.temperature.sharding == expected_sharding
    assert metadata.top_k.sharding == expected_sharding
    assert metadata.top_p.sharding == expected_sharding

    # 4. Check that values were correctly padded
    expected_temp = np.array(
        [
            0.7, 0.8, DEFAULT_SAMPLING_PARAMS["temperature"],
            DEFAULT_SAMPLING_PARAMS["temperature"]
        ],
        dtype=np.float32,
    )
    expected_top_k = np.array(
        [
            10, 20, DEFAULT_SAMPLING_PARAMS["top_k"],
            DEFAULT_SAMPLING_PARAMS["top_k"]
        ],
        dtype=np.int32,
    )
    expected_top_p = np.array(
        [
            0.9, 0.95, DEFAULT_SAMPLING_PARAMS["top_p"],
            DEFAULT_SAMPLING_PARAMS["top_p"]
        ],
        dtype=np.float32,
    )

    np.testing.assert_allclose(np.asarray(metadata.temperature), expected_temp)
    np.testing.assert_array_equal(np.asarray(metadata.top_k), expected_top_k)
    np.testing.assert_allclose(np.asarray(metadata.top_p), expected_top_p)


def test_from_input_batch_no_padding_needed(mesh: Mesh):
    """
    Tests the case where `num_reqs` equals `padded_num_reqs`, so **no padding** should occur.
    """
    num_reqs = 4
    padded_num_reqs = 4

    temp_tensor = np.array([0.7, 0.8, 0.6, 0.5], dtype=np.float32)
    top_k_tensor = np.array([10, 20, 30, 40], dtype=np.int32)
    top_p_tensor = np.array([0.9, 0.95, 0.85, 0.8], dtype=np.float32)

    mock_batch = MockInputBatch(
        all_greedy=False,
        num_reqs=num_reqs,
        temperature_cpu=temp_tensor,
        top_k_cpu=top_k_tensor,
        top_p_cpu=top_p_tensor,
    )

    metadata = TPUSupportedSamplingMetadata.from_input_batch(
        mesh=mesh, input_batch=mock_batch, padded_num_reqs=padded_num_reqs)

    assert metadata.do_sampling
    # Check that values are identical to the input, since no padding was needed
    np.testing.assert_allclose(np.asarray(metadata.temperature), temp_tensor)
    np.testing.assert_array_equal(np.asarray(metadata.top_k), top_k_tensor)
    np.testing.assert_allclose(np.asarray(metadata.top_p), top_p_tensor)


def test_jax_tree_util_registration():
    """
    Tests that the dataclass is correctly registered as a **JAX PyTree**,
    meaning `jax.tree_util` functions can operate on it as expected. 🌳
    """
    metadata = TPUSupportedSamplingMetadata(
        temperature=jnp.array([0.7]),
        top_k=jnp.array([10]),
        top_p=jnp.array([0.9]),
        do_sampling=True,
    )

    # Flatten the PyTree
    leaves, treedef = jax.tree_util.tree_flatten(metadata)

    # The leaves should be the "data_fields" specified in the decorator
    assert len(leaves) == 3
    np.testing.assert_array_equal(leaves[0], jnp.array([0.7]))
    np.testing.assert_array_equal(leaves[1], jnp.array([10]))
    np.testing.assert_array_equal(leaves[2], jnp.array([0.9]))

    # Reconstruct the PyTree from leaves
    new_metadata = jax.tree_util.tree_unflatten(treedef, leaves)

    # The reconstructed object should match the original
    assert new_metadata.do_sampling == metadata.do_sampling
    np.testing.assert_array_equal(new_metadata.temperature,
                                  metadata.temperature)
    np.testing.assert_array_equal(new_metadata.top_k, metadata.top_k)
    np.testing.assert_array_equal(new_metadata.top_p, metadata.top_p)


def test_from_input_batch_with_logprobs(mesh: Mesh):
    """
    Tests that the `logprobs` flag is correctly set based on `max_num_logprobs`.
    """
    # Case 1: Logprobs are requested
    mock_batch_with_logprobs = MockInputBatch(all_greedy=True,
                                              max_num_logprobs=5)
    metadata_with = TPUSupportedSamplingMetadata.from_input_batch(
        mesh=mesh,
        input_batch=mock_batch_with_logprobs,
        padded_num_reqs=4,
    )
    assert metadata_with.logprobs, "logprobs should be True when max_num_logprobs > 0"

    # Case 2: Logprobs are not requested (max_num_logprobs is 0)
    mock_batch_no_logprobs_zero = MockInputBatch(all_greedy=True,
                                                 max_num_logprobs=0)
    metadata_without_zero = TPUSupportedSamplingMetadata.from_input_batch(
        mesh=mesh,
        input_batch=mock_batch_no_logprobs_zero,
        padded_num_reqs=4,
    )
    assert not metadata_without_zero.logprobs, "logprobs should be False when max_num_logprobs is 0"

    # Case 3: Logprobs are not requested (max_num_logprobs is None)
    mock_batch_no_logprobs_none = MockInputBatch(all_greedy=True,
                                                 max_num_logprobs=None)
    metadata_without_none = TPUSupportedSamplingMetadata.from_input_batch(
        mesh=mesh,
        input_batch=mock_batch_no_logprobs_none,
        padded_num_reqs=4,
    )
    assert not metadata_without_none.logprobs, "logprobs should be False when max_num_logprobs is None"


def test_from_input_batch_sampling_with_logprobs(mesh: Mesh):
    """
    Tests enabling both sampling and logprobs simultaneously.
    """
    num_reqs = 2
    padded_num_reqs = 4
    mock_batch = MockInputBatch(
        all_greedy=False,
        num_reqs=num_reqs,
        temperature_cpu=np.zeros((padded_num_reqs, ), dtype=np.float32),
        top_k_cpu=np.zeros((padded_num_reqs, ), dtype=np.int32),
        top_p_cpu=np.zeros((padded_num_reqs, ), dtype=np.float32),
        max_num_logprobs=10,
    )

    metadata = TPUSupportedSamplingMetadata.from_input_batch(
        mesh=mesh, input_batch=mock_batch, padded_num_reqs=padded_num_reqs)

    assert metadata.do_sampling, "do_sampling should be True"
    assert metadata.logprobs, "logprobs should be True"



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import jax
import jax.numpy as jnp
import numpy as np
from flax import nnx
from jax.sharding import Mesh, PartitionSpec

from tpu_inference.layers.jax.moe.deepseek_v3_moe import DeepSeekV3Router
from tpu_inference.layers.jax.moe.moe import MoE, Router
from tpu_inference.layers.jax.moe.utils import MoEBackend


class TestDeepSeekV3Router(unittest.TestCase):

    def setUp(self):
        self.cpu_mesh = Mesh(jax.devices('cpu'), axis_names=('data', ))

    def test_get_topk_indices_single_group(self):
        """Test get_topk_indices with single expert group."""
        with jax.set_mesh(self.cpu_mesh):
            router = DeepSeekV3Router(random_init=True,
                                      hidden_size=512,
                                      num_experts=4,
                                      num_experts_per_tok=2,
                                      n_groups=1,
                                      topk_groups=1,
                                      norm_topk_prob=True,
                                      routed_scaling_factor=1.0,
                                      dtype=jnp.bfloat16,
                                      rngs=nnx.Rngs(42))
            router.bias_E = jnp.zeros((4, ))

            scores = jnp.array([[0.1, 0.3, 0.2, 0.4]])  # shape: (1, 4)
            indices = router.get_topk_indices(scores)

            # Should return indices of top 2 experts
            expected_indices = jnp.array([[3,
                                           1]])  # experts with scores 0.4, 0.3
            self.assertTrue(jnp.array_equal(indices, expected_indices))

    def test_get_topk_indices_2_groups(self):
        """Test get_topk_indices with 2 expert groups."""
        with jax.set_mesh(self.cpu_mesh):
            router = DeepSeekV3Router(random_init=True,
                                      hidden_size=512,
                                      num_experts=4,
                                      num_experts_per_tok=2,
                                      n_groups=2,
                                      topk_groups=1,
                                      norm_topk_prob=True,
                                      routed_scaling_factor=1.0,
                                      dtype=jnp.bfloat16,
                                      rngs=nnx.Rngs(42))
            router.bias_E = jnp.zeros((4, ))

            # 4 experts, 2 groups, 2 experts per group
            scores = jnp.array([[[0.1, 0.3, 0.2, 0.4]]])  # shape: (1, 1, 4)
            indices = router.get_topk_indices(scores)

            # Should return indices of top 2 experts
            expected_indices = jnp.array([[[3, 2]]])
            self.assertTrue(jnp.array_equal(indices, expected_indices))

    def test_router_e2e(self):
        with jax.set_mesh(self.cpu_mesh):
            router = DeepSeekV3Router(random_init=True,
                                      hidden_size=512,
                                      num_experts=8,
                                      num_experts_per_tok=2,
                                      n_groups=2,
                                      topk_groups=1,
                                      norm_topk_prob=True,
                                      routed_scaling_factor=1.0,
                                      dtype=jnp.bfloat16,
                                      rngs=nnx.Rngs(42))
            x = jnp.ones((2, 512))
            weights, indices = router(x)
            self.assertEqual(weights.shape, (2, 2))
            self.assertEqual(indices.shape, (2, 2))


class TestMoE(unittest.TestCase):

    def setUp(self):
        """Set up a multi-device mesh for testing."""
        devices = jax.devices()
        self.device_count = len(devices)
        if self.device_count < 8:
            self.skipTest("This test requires at least 8 simulated devices.")

        # This mesh will have a 'model' axis for expert parallelism
        mesh_shape = (self.device_count, 1)
        device_mesh_array = np.array(devices).reshape(mesh_shape)

        # Define the axis names
        axis_names = ('model', 'data')

        # Create the 2D mesh
        self.mesh = Mesh(device_mesh_array, axis_names=axis_names)

        # --- Model Configuration ---
        self.B, self.S, self.D = 2, 8, 32  # Batch, Sequence, Hidden Dim
        self.E, self.K = 8, 2  # Num Experts (Total), Experts per Token
        self.moe_intermediate_size = 64
        self.dtype = jnp.bfloat16
        self.key = jax.random.PRNGKey(42)

        # Input data
        self.x = jax.random.normal(self.key, (self.B * self.S, self.D),
                                   dtype=self.dtype)

    def _create_moe(self, backend: MoEBackend):
        """Helper to instantiate the MoE layer within the mesh context."""
        with self.mesh:
            router = Router(
                dtype=self.dtype,
                hidden_size=self.D,
                num_experts=self.E,
                num_experts_per_tok=self.K,
                router_act="softmax",  # Standard softmax routing
                rngs=nnx.Rngs(self.key),
                activation_ffw_td=PartitionSpec('data', 'model'),
                ed_sharding=PartitionSpec(None, 'model'),
                moe_backend=backend)

            moe = MoE(
                dtype=self.dtype,
                num_local_experts=self.E,
                hidden_size=self.D,
                intermediate_size_moe=self.moe_intermediate_size,
                hidden_act="silu",
                rngs=nnx.Rngs(self.key),
                router=router,
                mesh=self.mesh,

                # Sharding Specs
                activation_ffw_td=PartitionSpec('data', None),
                activation_ffw_ted=PartitionSpec('data', None),
                edf_sharding=PartitionSpec('model', None,
                                           None),  # Expert par on axis 0
                efd_sharding=PartitionSpec('model', None, None),
                apply_expert_weight_before_computation=False,
                moe_backend=backend,
                num_experts_per_tok=self.K,
            )
        return moe

    def _compute_ground_truth(self, moe_instance, x_input):
        """
        Computes the expected MoE output using a simple, sequential Python loop.
        This serves as the 'Gold Standard' to verify distributed logic.
        """
        # 1. Get Router Decisions
        # We run the router (which is replicated/simple) to get weights & indices
        weights, indices = moe_instance.router(x_input)

        # 2. Fetch Full Weights from Devices
        # The MoE weights are sharded. We fetch them to CPU/Host as full arrays.
        gating_kernel_full = jax.device_get(
            moe_instance.kernel_gating_EDF.value)
        up_proj_kernel_full = jax.device_get(
            moe_instance.kernel_up_proj_EDF.value)
        down_proj_kernel_full = jax.device_get(
            moe_instance.kernel_down_proj_EFD.value)

        # 3. Flatten for iteration
        flat_x = x_input.reshape(-1, self.D)
        flat_weights = weights.reshape(-1, self.K)
        flat_indices = indices.reshape(-1, self.K)

        expected_output = np.zeros_like(flat_x)

        # 4. Sequential computation per token per selected expert
        for t in range(flat_x.shape[0]):
            token_val = flat_x[t]
            token_accum = np.zeros_like(token_val)

            for k in range(self.K):
                expert_idx = flat_indices[t, k]
                router_weight = flat_weights[t, k]

                # Extract specific expert weights
                w_gating = gating_kernel_full[expert_idx]
                w_up = up_proj_kernel_full[expert_idx]
                w_down = down_proj_kernel_full[expert_idx]

                # Forward pass: Up(x) * SiLU(Gate(x))
                gate_out = np.dot(token_val, w_gating)
                up_out = np.dot(token_val, w_up)

                # SiLU activation
                # silu(x) = x * sigmoid(x)
                silu_out = gate_out * (1 / (1 + np.exp(-gate_out)))

                expert_out = silu_out * up_out

                # Down projection
                down_out = np.dot(expert_out, w_down)

                # Weighted sum
                token_accum += down_out * router_weight

            expected_output[t] = token_accum

        return jnp.array(expected_output, dtype=self.dtype)

    def test_dense_backend_correctness(self):
        """Verifies the DENSE_MAT backend against the sequential ground truth."""
        moe = self._create_moe(MoEBackend.DENSE_MAT)

        with self.mesh:
            actual_output = moe(self.x)

        expected_output = self._compute_ground_truth(moe, self.x)

        # Dense matmul should be very numerically close
        self.assertTrue(
            jnp.allclose(actual_output, expected_output, atol=1e-2, rtol=1e-2),
            "Dense backend output does not match ground truth.")

    def test_sparse_distributed_backend_correctness_ragged_dot(self):
        """
        Verifies the RAGGED_DOT (Sparse) backend with expert parallelism
        against the sequential ground truth.
        """
        # 1. Instantiate MoE with RAGGED_DOT backend
        # TODO: add MoEBackend.FUSED_MOE, MoEBackend.VLLM_MOE
        for backend in [MoEBackend.RAGGED_DOT, MoEBackend.MEGABLX_GMM]:
            moe = self._create_moe(backend)

            # 2. Run Forward Pass (Distributed)
            # The __call__ invokes shard_map internally for RAGGED_DOT
            with self.mesh:
                actual_output = moe(self.x)

            # 3. Compute Ground Truth using the exact same weights
            # (Weights are pulled from the initialized model)
            expected_output = self._compute_ground_truth(moe, self.x)

            # 4. Compare
            # Due to permutation re-ordering and potential float precision diffs
            # in different kernels, we allow a small tolerance.
            diff = jnp.mean(jnp.abs(actual_output - expected_output))

            self.assertTrue(
                jnp.allclose(actual_output,
                             expected_output,
                             atol=5e-2,
                             rtol=5e-2),
                f"Sparse distributed output mismatch for backebd tyoe {backend}. Mean diff: {diff}"
            )

    def test_backend_consistency(self):
        """
        Ensures that if we initialize two models with identical seeds/weights,
        Dense and Sparse backends yield the same result.
        """
        # This test requires careful weight synchronization or fixed seeds.
        # Since nnx.Rngs(key) is deterministic, we just re-use the key.

        # 1. Run Dense
        moe_dense = self._create_moe(MoEBackend.DENSE_MAT)
        with self.mesh:
            out_dense = moe_dense(self.x)

        # 2. Run Sparse
        # We must re-init with same key to get same weights
        moe_sparse = self._create_moe(MoEBackend.MEGABLX_GMM)
        with self.mesh:
            out_sparse = moe_sparse(self.x)

        self.assertTrue(
            jnp.allclose(out_dense, out_sparse, atol=5e-2, rtol=5e-2),
            "Dense and Sparse backends produced different results for identical initialization."
        )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from abc import ABC, abstractmethod

import jax
from vllm.config import VllmConfig

from tpu_inference.layers.jax import JaxModule


def get_tpu_quantization_config(vllm_config: VllmConfig):
    from tpu_inference.layers.jax.quantization.unquantized import \
        UnquantizedConfig

    model_config = copy.deepcopy(vllm_config.model_config)
    method_to_config: dict[str | None, type] = {
        None: UnquantizedConfig,
    }
    if model_config.quantization not in method_to_config:
        raise NotImplementedError(
            f"{model_config.quantization} quantization method not supported."
            f" Supported methods are {method_to_config.keys()}")
    quant_config = method_to_config[model_config.quantization]
    return quant_config()


class QuantizeMethodBase(ABC):
    """Base class for different quantized methods."""

    def create_weights_jax(self, layer: JaxModule, *weight_args,
                           **extra_weight_attrs):
        """Create weights for a layer.

        The weights will be set as attributes of the layer."""
        pass

    @abstractmethod
    def apply_jax(self, layer: JaxModule, *args, **kwargs) -> jax.Array:
        """Apply the weights in layer to the input tensor.

        Expects create_weights to have been called before on the layer."""
        raise NotImplementedError



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project
import os
import time
from dataclasses import asdict

import pytest
from vllm import LLM, EngineArgs, SamplingParams


@pytest.fixture
def model_name():
    """Choose gemma-27b as the test model as it has both full attention and 
    sliding window attention."""
    return "google/gemma-3-27b-it"


@pytest.fixture
def test_prompts():
    """Simple test prompts for hybrid kv cache testing."""
    return [
        "Hello, my name is",
        "The capital of France is",
        "The colors of the rainbow are",
        "The future of AI is",
        "The president of the United States is",
        "How many players are on a standard soccer team?",
        "In Greek mythology, who is the god of the sea?",
        "What is the capital of Australia?",
        "What is the largest planet in our solar system?",
        "Who developed the theory of general relativity?",
    ]


@pytest.fixture
def sampling_params():
    """Standard sampling parameters for testing."""
    return SamplingParams(
        temperature=0.0,
        max_tokens=32,
        ignore_eos=True,
        logprobs=1,
    )


def _run_inference_with_config(
        model_name: str,
        test_prompts: list,
        sampling_params: SamplingParams,
        tensor_parallel_size: int = 4,
        kv_cache_dtype: str = "auto",
        enable_prefix_caching: bool = False,
        disable_hybrid_kv_cache_manager: bool = False) -> list:
    """Helper function to run inference with specified configuration."""

    # Create LLM args using parser-based approach similar to offline_inference.py
    engine_args = EngineArgs(
        model=model_name,
        max_model_len=64,
        tensor_parallel_size=tensor_parallel_size,
        gpu_memory_utilization=0.95,
        max_num_batched_tokens=256,
        max_num_seqs=16,
        enable_prefix_caching=enable_prefix_caching,
        kv_cache_dtype=kv_cache_dtype,
        disable_hybrid_kv_cache_manager=disable_hybrid_kv_cache_manager,
    )

    engine_args_dict = asdict(engine_args)
    llm = LLM(**engine_args_dict)

    try:
        outputs = llm.generate(test_prompts, sampling_params)
        return outputs
    finally:
        del llm
        # Wait for TPUs to be released
        time.sleep(10)


def test_hybrid_kv_cache(
    model_name: str,
    test_prompts: list,
    sampling_params: SamplingParams,
):
    """
    Test hybrid kv cache works on gemma vLLM models.
    """

    os.environ['MODEL_IMPL_TYPE'] = 'vllm'
    # Test with hybrid kv cache alloctaion enabled.
    outputs = _run_inference_with_config(
        model_name=model_name,
        test_prompts=test_prompts,
        sampling_params=sampling_params,
        disable_hybrid_kv_cache_manager=False,
    )

    # Verify we got outputs for all prompts
    assert len(outputs) == len(test_prompts)

    # Verify each output has generated text
    for output in outputs:
        assert len(output.outputs) > 0
        assert len(output.outputs[0].text.strip()) > 0

    print(f"✓ Hybrid KV cache test passed with {len(outputs)} outputs")


def test_hybrid_kv_cache_correctness(
    model_name: str,
    test_prompts: list,
    sampling_params: SamplingParams,
):
    """
    Test that hybrid kv cache allocation produces consistent results compared 
    to standard kv cache allocation.
    """
    os.environ['SKIP_JAX_PRECOMPILE'] = '1'
    os.environ['VLLM_XLA_CHECK_RECOMPILATION'] = '0'

    small_prompts = test_prompts

    # Run baseline (no hybrid kv cache)
    baseline_outputs = _run_inference_with_config(
        model_name=model_name,
        test_prompts=small_prompts,
        sampling_params=sampling_params,
        disable_hybrid_kv_cache_manager=True,
    )

    # Run with hybrid kv cache enabled.
    hybrid_kvcache_outputs = _run_inference_with_config(
        model_name=model_name,
        test_prompts=small_prompts,
        sampling_params=sampling_params,
        disable_hybrid_kv_cache_manager=False,
    )

    # Compare outputs - in theory they should be identical for greedy sampling
    # in reality there may be some differences, but overall the outputs should
    # be very similar.

    # an example:
    # prompt: What is the capital of Australia?
    # both answers should be acceptable.
    # The capital of Australia is Canberra. It is located in the Australian Capital Territory (ACT) and is home to many
    # Canberra is the capital of Australia. It is located in the Australian Capital Territory (ACT) and is home to
    assert len(baseline_outputs) == len(hybrid_kvcache_outputs)

    text_matches = 0
    text_mismatches = 0
    logprob_mismatches = 0
    max_logprob_diff = 0.0

    for i, (baseline, hybrid_kvcache_result) in enumerate(
            zip(baseline_outputs, hybrid_kvcache_outputs)):
        baseline_text = baseline.outputs[0].text.strip()
        hybrid_kvcache_text = hybrid_kvcache_result.outputs[0].text.strip()

        # Check text output
        if baseline_text == hybrid_kvcache_text:
            text_matches += 1
        else:
            text_mismatches += 1
            print(f"Text mismatch found in prompt {i}:")
            print(f"  Baseline: {baseline_text}")
            print(f"  Hybrid KV Cache: {hybrid_kvcache_text}")

        # Check log probabilities
        baseline_logprobs = baseline.outputs[0].logprobs
        hybrid_kvcache_logprobs = hybrid_kvcache_result.outputs[0].logprobs
        if baseline_logprobs is not None and hybrid_kvcache_logprobs is not None:
            # Compare log probabilities for each token
            assert len(baseline_logprobs) == len(hybrid_kvcache_logprobs), \
                f"Logprobs length mismatch: {len(baseline_logprobs)} vs {len(hybrid_kvcache_logprobs)}"
            for token_idx, (base_lp, hybrid_kvcache_lp) in enumerate(
                    zip(baseline_logprobs, hybrid_kvcache_logprobs)):
                # Get the top logprob value for the selected token
                if base_lp and hybrid_kvcache_lp:
                    # Get the top token's logprob from each
                    base_top_token = list(base_lp.keys())[0]
                    hybrid_kvcache_top_token = list(
                        hybrid_kvcache_lp.keys())[0]

                    base_logprob_val = base_lp[base_top_token].logprob
                    hybrid_kvcache_logprob_val = hybrid_kvcache_lp[
                        hybrid_kvcache_top_token].logprob

                    # Calculate absolute difference
                    diff = abs(base_logprob_val - hybrid_kvcache_logprob_val)
                    max_logprob_diff = max(max_logprob_diff, diff)

                    # Allow small numerical differences (e.g., 1e-3)
                    if diff > 1e-3:
                        logprob_mismatches += 1
                        print(
                            f"Logprob mismatch in prompt {i}, token {token_idx}:"
                        )
                        print(
                            f"  Baseline token: {base_top_token}, logprob: {base_logprob_val:.6f}"
                        )
                        print(
                            f"  Hybrid KV Cache token: {hybrid_kvcache_top_token}, logprob: {hybrid_kvcache_logprob_val:.6f}"
                        )
                        print(f"  Difference: {diff:.6f}")

    print("✓ Correctness test results:")
    print(f"  Text: {text_matches} matches, {text_mismatches} mismatches")
    print(f"  Max logprob difference: {max_logprob_diff:.6e}")
    print(f"  Significant logprob mismatches (>1e-3): {logprob_mismatches}")

    # Allow for some variance due to potential numerical differences
    # but most outputs should match with greedy sampling
    text_match_rate = text_matches / len(baseline_outputs)
    assert text_match_rate >= 0.9, f"Text match rate {text_match_rate:.2%} is too low"

    # Log probabilities should be very close (allow small numerical errors)
    assert max_logprob_diff < 2, f"Max logprob difference {max_logprob_diff} is too large"



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import InitVar, dataclass
from typing import Tuple

import jax
import jax.numpy as jnp
from flax import nnx
from flax.typing import Sharding
from jax.sharding import Mesh
from jax.sharding import PartitionSpec as P

from tpu_inference import utils
from tpu_inference.kernels.ragged_paged_attention.v3.kernel_hd64 import \
    ragged_paged_attention_hd64
from tpu_inference.layers.common.attention_metadata import AttentionMetadata
from tpu_inference.layers.common.quantization import quantize_kv
from tpu_inference.layers.jax.base import create_param
from tpu_inference.layers.jax.rope import GptOssRotaryEmbedding

KVCache = Tuple[jax.Array, jax.Array]


@dataclass(kw_only=True)
class GptOssAttention(nnx.Module):
    """
    JAX implementation of the GPT-OSS Attention block
    """
    hidden_size: int
    num_attention_heads: int
    num_key_value_heads: int
    head_dim: int
    dtype: jnp.dtype
    rngs: InitVar[nnx.Rngs]

    rope_theta: float
    initial_context_length: int = 4096
    rope_scaling_factor: float = 32.0
    rope_ntk_alpha: float = 1.0
    rope_ntk_beta: float = 32.0
    kv_cache_dtype: str

    query_tnh: P = P()
    keyvalue_skh: P = P()
    attn_o_tnh: P = P()
    dnh_sharding: Sharding = ()
    dkh_sharding: Sharding = ()
    nhd_sharding: Sharding = ()
    n_sharding: Sharding = ()
    nh_sharding: Sharding = ()
    kh_sharding: Sharding = ()
    d_sharding: Sharding = ()

    random_init: bool = False
    mesh: Mesh

    _q_scale: float = 1.0
    _k_scale: float = 1.0
    _v_scale: float = 1.0
    kv_cache_quantized_dtype = None

    def __post_init__(self, rngs: nnx.Rngs):
        """Initializes weights, biases, and RoPE module."""

        self.sm_scale = 1.0 / (self.head_dim**0.5)

        self.sinks_N = create_param(
            rngs,
            shape=(self.num_attention_heads, ),
            dtype=jnp.float32,
            sharding=self.n_sharding,
            random_init=self.random_init,
        )

        # Q, K, V projection kernels
        self.kernel_q_DNH = create_param(
            rngs,
            shape=(self.hidden_size, self.num_attention_heads, self.head_dim),
            dtype=self.dtype,
            sharding=self.dnh_sharding,
            random_init=self.random_init,
        )
        self.bias_q_NH = create_param(
            rngs,
            shape=(self.num_attention_heads, self.head_dim),
            dtype=self.dtype,
            sharding=self.nh_sharding,
            random_init=self.random_init,
        )
        self.kernel_k_DKH = create_param(
            rngs,
            shape=(self.hidden_size, self.num_key_value_heads, self.head_dim),
            dtype=self.dtype,
            sharding=self.dkh_sharding,
            random_init=self.random_init,
        )
        self.bias_k_KH = create_param(
            rngs,
            shape=(self.num_key_value_heads, self.head_dim),
            dtype=self.dtype,
            sharding=self.kh_sharding,
            random_init=self.random_init,
        )
        self.kernel_v_DKH = create_param(
            rngs,
            shape=(self.hidden_size, self.num_key_value_heads, self.head_dim),
            dtype=self.dtype,
            sharding=self.dkh_sharding,
            random_init=self.random_init,
        )
        self.bias_v_KH = create_param(
            rngs,
            shape=(self.num_key_value_heads, self.head_dim),
            dtype=self.dtype,
            sharding=self.kh_sharding,
            random_init=self.random_init,
        )
        # Output projection kernel
        self.kernel_o_proj_NHD = create_param(
            rngs,
            shape=(self.num_attention_heads, self.head_dim, self.hidden_size),
            dtype=self.dtype,
            sharding=self.nhd_sharding,
            random_init=self.random_init,
        )
        self.bias_o_D = create_param(
            rngs,
            shape=(self.hidden_size, ),
            dtype=self.dtype,
            sharding=self.d_sharding,
            random_init=self.random_init,
        )

        # RoPE Module
        self.rope = GptOssRotaryEmbedding(
            head_dim=self.head_dim,
            rope_theta=self.rope_theta,
            dtype=self.dtype,
            initial_context_length=self.initial_context_length,
            rope_scaling_factor=self.rope_scaling_factor,
            rope_ntk_alpha=self.rope_ntk_alpha,
            rope_ntk_beta=self.rope_ntk_beta)

        if self.kv_cache_dtype != "auto":
            self.kv_cache_quantized_dtype = utils.get_jax_dtype_from_str_dtype(
                self.kv_cache_dtype)

    def attention(
        self,
        kv_cache: KVCache,
        q_TNH: jax.Array,
        k_SKH: jax.Array,
        v_SKH: jax.Array,
        sinks: jax.Array,
        attention_metadata: AttentionMetadata,
        mesh: Mesh,
        q_scale: float | None = None,
        k_scale: float | None = None,
        v_scale: float | None = None,
    ) -> Tuple[KVCache, jax.Array]:
        """Performs scaled dot-product attention by calling the ragged_paged_attention kernel."""
        md = attention_metadata
        kv_cache_spec = P("data", None, "model")

        in_specs = (
            self.query_tnh,  # q
            self.keyvalue_skh,  # k
            self.keyvalue_skh,  # v
            kv_cache_spec,  # kv_cache
            P("data"),  # md.seq_lens
            P("data"),  # page_indices_flat
            P("data"),  # query_start_loc
            P("data"),  # distribution
            P(('model')),  # sinks
        )
        out_specs = (self.attn_o_tnh, kv_cache_spec)

        def _ragged_paged_attention_wrapper(*args):
            # Pass the GPT-OSS specific parameters to the kernel
            return ragged_paged_attention_hd64(
                *args,
                sm_scale=self.sm_scale,
                sliding_window=md.sliding_window,
                q_scale=q_scale,
                k_scale=k_scale,
                v_scale=v_scale,
            )

        output_TNH, kv_cache = jax.jit(
            jax.shard_map(
                _ragged_paged_attention_wrapper,
                mesh=mesh,
                in_specs=in_specs,
                out_specs=out_specs,
                check_vma=False,
            ))(
                q_TNH,
                k_SKH,
                v_SKH,
                kv_cache,
                md.seq_lens,
                md.block_tables,
                md.query_start_loc,
                md.request_distribution,
                sinks,
            )
        return kv_cache, output_TNH

    def __call__(self,
                 x_TD,
                 is_prefill,
                 kv_cache: KVCache,
                 attention_metadata: AttentionMetadata,
                 use_attention_rope: bool = True):
        """Forward pass for the Attention module using 3D kernels."""
        md = attention_metadata
        x_TD = jnp.asarray(x_TD, self.dtype)

        with jax.named_scope("q_proj"):
            q_TNH = jnp.einsum("TD,DNH->TNH", x_TD, self.kernel_q_DNH.value)
            q_TNH += self.bias_q_NH.value

        with jax.named_scope("k_proj"):
            k_TKH = jnp.einsum("TD,DKH->TKH", x_TD, self.kernel_k_DKH.value)
            k_TKH += self.bias_k_KH.value

        with jax.named_scope("v_proj"):
            v_TKH = jnp.einsum("TD,DKH->TKH", x_TD, self.kernel_v_DKH.value)
            v_TKH += self.bias_v_KH.value

        if use_attention_rope:
            q_TNH, k_TKH = self.rope(q_TNH, k_TKH, md.input_positions)

        q_scale = k_scale = v_scale = None
        if self.kv_cache_quantized_dtype:
            # TODO(kyuyeunk/jacobplatin): Enable w8a8 when VREG spill issue is resolved.
            # q_scale = self._q_scale
            k_scale = self._k_scale
            v_scale = self._v_scale
            k_TKH, v_TKH = quantize_kv(self.kv_cache_quantized_dtype, k_TKH,
                                       v_TKH, k_scale, v_scale)

        with jax.named_scope("attn_op"):
            new_kv_cache, attn_out_TNH = self.attention(
                kv_cache=kv_cache,
                q_TNH=q_TNH,
                k_SKH=k_TKH,
                v_SKH=v_TKH,
                sinks=self.sinks_N.value,
                attention_metadata=md,
                mesh=self.mesh,
                q_scale=q_scale,
                k_scale=k_scale,
                v_scale=v_scale,
            )
            attn_out_TNH = attn_out_TNH[..., :self.head_dim]

        with jax.named_scope("o_proj"):
            output_TD = jnp.einsum("TNH,NHD->TD", attn_out_TNH,
                                   self.kernel_o_proj_NHD.value)
            output_TD += self.bias_o_D.value

        return new_kv_cache, output_TD



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock

import jax
import jax.numpy as jnp
import numpy as np
import pytest
from jax.sharding import Mesh

from tpu_inference.layers.common.attention_interface import (
    attention, sharded_ragged_paged_attention)
from tpu_inference.layers.common.attention_metadata import AttentionMetadata
from tpu_inference.layers.common.sharding import ShardingAxisName
from tpu_inference.runner.kv_cache import get_kv_cache_shape_with_mesh

# ---- Test Configuration & Constants ----

# Total number of tokens across all sequences in the batch
TOTAL_TOKENS = 10
# Number of sequences in the batch
NUM_SEQS = 2
# Padded maximum number of sequences
MAX_NUM_SEQS = 4
# Number of attention heads (Query)
NUM_HEADS = 8
# Number of attention heads (Key/Value) - for Grouped-Query Attention
NUM_KV_HEADS = 4
# Total number of blocks in the KV cache
NUM_BLOCKS = 32
# Number of tokens per block
BLOCK_SIZE = 16
# Maximum number of blocks a single sequence can occupy
MAX_BLOCKS_PER_SEQ = 8


@pytest.fixture
def mesh():
    """Provides a mock 1D JAX mesh for testing."""
    # Create a mesh with available devices, useful for running on CPU/GPU/TPU
    # For this test, it will likely be a single CPU device.
    devices = np.array(jax.local_devices()[:1])
    if not devices.any():
        # Add a mock device if no devices are present (e.g., in a CI environment)
        devices = np.array([jax.devices("cpu")[0]])
    return Mesh(devices.reshape((-1, 1, 1)), ("data", "attn_dp", "model"))


# ---- Test for `attention` ----


def _test_attention(monkeypatch, mesh, head_dim, use_sinks=False):
    """
    Tests the main `attention` function.

    Verifies that:
    1. It calls the `sharded_ragged_paged_attention` kernel with correct metadata.
    2. The final outputs (kv_cache and attention output) have the correct shapes.
    """
    # 1. Arrange

    # Create input tensors
    q_dtype = jnp.float32
    kv_dtype = jnp.float32
    q = jnp.ones((TOTAL_TOKENS, NUM_HEADS, head_dim), dtype=q_dtype)
    k = jnp.ones((TOTAL_TOKENS, NUM_KV_HEADS, head_dim), dtype=kv_dtype)
    v = jnp.ones((TOTAL_TOKENS, NUM_KV_HEADS, head_dim), dtype=kv_dtype)
    sinks = jnp.ones((NUM_HEADS, ), dtype=jnp.float32) if use_sinks else None

    kv_cache_shape = get_kv_cache_shape_with_mesh(
        mesh,
        NUM_BLOCKS,
        BLOCK_SIZE,
        NUM_KV_HEADS,
        head_dim,
        kv_dtype,
    )
    kv_cache = jnp.zeros(kv_cache_shape, dtype=kv_dtype)

    # Mock ragged_paged_attention to return a tensor of the correct shape
    mock_paged_attn_kernel = MagicMock(return_value=(jnp.ones(
        (TOTAL_TOKENS, NUM_HEADS, head_dim)), kv_cache), )

    if head_dim == 64:
        monkeypatch.setattr(
            "tpu_inference.layers.common.attention_interface.ragged_paged_attention_hd64",
            mock_paged_attn_kernel,
        )
    else:
        monkeypatch.setattr(
            "tpu_inference.layers.common.attention_interface.ragged_paged_attention",
            mock_paged_attn_kernel,
        )

    # Create AttentionMetadata
    attention_metadata = AttentionMetadata(
        input_positions=jnp.arange(TOTAL_TOKENS, dtype=jnp.int32),
        block_tables=jnp.zeros((MAX_NUM_SEQS * MAX_BLOCKS_PER_SEQ, ),
                               dtype=jnp.int32),
        seq_lens=jnp.array([5, 5, 0, 0], dtype=jnp.int32),
        query_start_loc=jnp.array([0, 5, 10, 10, 10], dtype=jnp.int32),
        request_distribution=jnp.array([0, 0, NUM_SEQS], dtype=jnp.int32),
    )

    # 2. Act
    final_kv_cache, output = attention(
        kv_cache=kv_cache,
        q=q,
        k=k,
        v=v,
        attention_metadata=attention_metadata,
        mesh=mesh,
        head_dim_original=head_dim,
        sinks=sinks,
    )

    # 3. Assert
    # Check that both mocked kernels were called
    mock_paged_attn_kernel.assert_called_once()

    # Check output shapes
    assert final_kv_cache.shape == kv_cache.shape
    assert output.shape == q.shape

    # Check that the output is the one from our mock
    assert jnp.all(output == 1.0)


def test_attention(monkeypatch, mesh):
    _test_attention(monkeypatch, mesh, 128)


def test_attention_hd64(monkeypatch, mesh):
    _test_attention(monkeypatch, mesh, 64)


def test_attention_sink(monkeypatch, mesh):
    _test_attention(monkeypatch, mesh, 64, True)


def test_attention_sink_no_64_raises_error(monkeypatch, mesh):
    with pytest.raises(
            NotImplementedError,
            match="Attention sink support is only available when head_dim==64"
    ):
        _test_attention(monkeypatch, mesh, 128, True)


# ---- Tests for `sharded_ragged_paged_attention` ----


@pytest.fixture
def gqa_mesh():
    """Provides a mock JAX mesh for GQA testing with tensor parallelism."""
    # This mesh has 8 devices for tensor parallelism over heads.
    # We create a 1x8 mesh for ('attn_data', 'attn_head')
    try:
        devices = np.array(jax.local_devices()[:1] * 4)
        if devices.size == 0:
            raise IndexError
    except IndexError:
        # Fails in environments with no devices
        devices = np.array([jax.devices("cpu")[0]] * 4)

    return Mesh(
        devices.reshape((1, 4)),
        (
            ShardingAxisName.ATTN_DATA,
            ShardingAxisName.ATTN_HEAD,
        ),
    )


def test_sharded_ragged_paged_attention_gqa_replication(monkeypatch, gqa_mesh):
    """
    Tests that K and V heads are correctly replicated for GQA in
    `sharded_ragged_paged_attention`.
    """
    # 1. Arrange
    tp_size = gqa_mesh.shape[ShardingAxisName.ATTN_HEAD]
    assert tp_size == 4
    num_kv_heads = 2  # num_kv_heads < tp_size and tp_size % num_kv_heads == 0
    head_dim = 128
    factor = tp_size // num_kv_heads

    q = jnp.ones((TOTAL_TOKENS, NUM_HEADS, head_dim))
    # Create K and V with values that can be checked after repeating
    k_content = jnp.arange(TOTAL_TOKENS * num_kv_heads * head_dim).reshape(
        (TOTAL_TOKENS, num_kv_heads, head_dim))
    v_content = -k_content
    k = k_content
    v = v_content

    # The actual shape of kv_cache does not matter as much since we mock the call
    kv_cache = jnp.zeros((num_kv_heads, NUM_BLOCKS, BLOCK_SIZE, head_dim))

    # Other metadata, can be zero/empty for this test's purpose
    kv_lens = jnp.zeros((MAX_NUM_SEQS, ), dtype=jnp.int32)
    page_indices = jnp.zeros((MAX_NUM_SEQS, MAX_BLOCKS_PER_SEQ),
                             dtype=jnp.int32)
    cu_q_lens = jnp.zeros((MAX_NUM_SEQS + 1, ), dtype=jnp.int32)
    distribution = jnp.zeros((3, ), dtype=jnp.int32)
    sm_scale = 1.0

    # Mock jax.shard_map to capture the arguments passed to its mapped function
    mock_shard_map_callable = MagicMock(return_value=(jnp.ones_like(q),
                                                      kv_cache))
    mock_shard_map = MagicMock(return_value=mock_shard_map_callable)
    monkeypatch.setattr("jax.shard_map", mock_shard_map)

    # 2. Act
    sharded_ragged_paged_attention(
        mesh=gqa_mesh,
        q=q,
        k=k,
        v=v,
        kv_cache=kv_cache,
        kv_lens=kv_lens,
        page_indices=page_indices,
        cu_q_lens=cu_q_lens,
        distribution=distribution,
        attention_sink=None,
        sm_scale=sm_scale,
    )

    # 3. Assert
    # Check that shard_map was called
    mock_shard_map.assert_called_once()
    # Check that the function returned by shard_map was called with arguments
    mock_shard_map_callable.assert_called_once()

    # Get the arguments passed to the jitted function inside shard_map
    call_args = mock_shard_map_callable.call_args[0]
    replicated_k = call_args[1]
    replicated_v = call_args[2]

    # Check shapes
    assert replicated_k.shape[1] == tp_size
    assert replicated_v.shape[1] == tp_size
    assert replicated_k.shape[1] == k.shape[1] * factor
    assert replicated_v.shape[1] == v.shape[1] * factor

    # Check content of replicated K
    expected_k = jnp.repeat(k_content, factor, axis=1)
    assert jnp.array_equal(replicated_k, expected_k)

    # Check content of replicated V
    expected_v = jnp.repeat(v_content, factor, axis=1)
    assert jnp.array_equal(replicated_v, expected_v)


def test_sharded_ragged_paged_attention_gqa_incompatible_raises_error(
    gqa_mesh, ):
    """
    Tests that a ValueError is raised for GQA when tp_size is not divisible
    by num_kv_heads.
    """
    # 1. Arrange
    tp_size = gqa_mesh.shape[ShardingAxisName.ATTN_HEAD]
    assert tp_size == 4
    num_kv_heads = 3  # Incompatible with tp_size=4
    head_dim = 128

    q = jnp.ones((TOTAL_TOKENS, NUM_HEADS, head_dim))
    k = jnp.ones((TOTAL_TOKENS, num_kv_heads, head_dim))
    v = jnp.ones((TOTAL_TOKENS, num_kv_heads, head_dim))
    kv_cache = jnp.zeros((num_kv_heads, NUM_BLOCKS, BLOCK_SIZE, head_dim))
    # Other metadata
    kv_lens = jnp.zeros((MAX_NUM_SEQS, ), dtype=jnp.int32)
    page_indices = jnp.zeros((MAX_NUM_SEQS, MAX_BLOCKS_PER_SEQ),
                             dtype=jnp.int32)
    cu_q_lens = jnp.zeros((MAX_NUM_SEQS + 1, ), dtype=jnp.int32)
    distribution = jnp.zeros((3, ), dtype=jnp.int32)
    sm_scale = 1.0

    # 2. Act & Assert
    with pytest.raises(
            ValueError,
            match=(f"For GQA/MQA, tp_size {tp_size} must be divisible by "
                   f"num_kv_heads {num_kv_heads}"),
    ):
        sharded_ragged_paged_attention(
            mesh=gqa_mesh,
            q=q,
            k=k,
            v=v,
            kv_cache=kv_cache,
            kv_lens=kv_lens,
            page_indices=page_indices,
            cu_q_lens=cu_q_lens,
            distribution=distribution,
            attention_sink=None,
            sm_scale=sm_scale,
        )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Protocol

from flax import nnx
from vllm.distributed.utils import get_pp_indices

from tpu_inference.distributed.jax_parallel_state import get_pp_group


class PPMissingLayer(nnx.Module):
    """
    A placeholder layer for missing layers in a pipeline parallel model.
    """

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        """Return the first arg from args or the first value from kwargs."""
        return args[0] if args else next(iter(kwargs.values()))


class LayerFn(Protocol):

    def __call__(self) -> nnx.Module:
        ...


def make_layers(
    num_hidden_layers: int,
    layer_fn: LayerFn,
) -> tuple[int, int, List[nnx.Module]]:
    start_layer, end_layer = get_pp_indices(num_hidden_layers,
                                            get_pp_group().rank_in_group,
                                            get_pp_group().world_size)

    layers = [PPMissingLayer() for _ in range(start_layer)] \
        + [layer_fn() for _ in range(start_layer, end_layer)] \
        + [PPMissingLayer() for _ in range(end_layer, num_hidden_layers)]

    return start_layer, end_layer, layers
